# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
#  under the License.

import mock
import netaddr
from oslo_log import log as logging

from quark.db import api as db_api
from quark.db import models
from quark import tags
from quark.tests.functional.base import BaseFunctionalTest

LOG = logging.getLogger(__name__)


class TestDBAPI(BaseFunctionalTest):
    def setUp(self):
        super(TestDBAPI, self).setUp()

    def test_port_find_ip_address_id(self):
        self.context.session.query = mock.Mock()
        db_api.port_find(self.context, ip_address_id="fake")
        query_obj = self.context.session.query.return_value
        filter_fn = query_obj.options.return_value.filter
        self.assertEqual(filter_fn.call_count, 1)

    def test_ip_address_find_device_id(self):
        query_mock = mock.Mock()
        filter_mock = mock.Mock()

        self.context.session.query = query_mock
        query_mock.return_value = filter_mock

        db_api.ip_address_find(self.context, device_id="foo")
        self.assertEqual(filter_mock.filter.call_count, 1)

    def test_ip_address_find_address_type(self):
        self.context.session.query = mock.MagicMock()
        filter_mock = self.context.session.query.return_value

        db_api.ip_address_find(self.context, address_type="foo")
        # NOTE(thomasem): Creates sqlalchemy.sql.elements.BinaryExpression
        # when using SQLAlchemy models in expressions.
        tenant_filter = (models.IPAddress.used_by_tenant_id.in_(
                         [self.context.tenant_id]))
        type_filter = models.IPAddress.address_type == "foo"
        self.assertEqual(len(filter_mock.filter.call_args[0]), 2)
        # NOTE(thomasem): Unfortunately BinaryExpression.compare isn't
        # showing to be a reliable comparison, so using the string
        # representation which dumps the associated SQL for the filter.
        type_found = False
        tenant_found = False
        for model_filter in list(filter_mock.filter.call_args[0]):
            if str(type_filter) == str(model_filter):
                type_found = True
            elif str(tenant_filter) == str(model_filter):
                tenant_found = True
        self.assertTrue(tenant_found)
        self.assertTrue(type_found)

    def test_ip_address_find_port_id(self):
        self.context.session.query = mock.MagicMock()
        final_query_mock = self.context.session.query.return_value

        db_api.ip_address_find(self.context, port_id="foo")
        # NOTE(thomasem): Creates sqlalchemy.sql.elements.BinaryExpression
        # when using SQLAlchemy models in expressions.
        tenant_filter = (models.IPAddress.used_by_tenant_id.in_(
                         [self.context.tenant_id]))
        port_filter = models.IPAddress.ports.any(models.Port.id == "foo")
        self.assertEqual(len(final_query_mock.filter.call_args[0]), 2)
        port_found = False
        tenant_found = False
        for model_filter in list(final_query_mock.filter.call_args[0]):
            if str(port_filter) == str(model_filter):
                port_found = True
            elif str(tenant_filter) == str(model_filter):
                tenant_found = True
        self.assertTrue(tenant_found)
        self.assertTrue(port_found)

    def test_ip_address_find_address_type_as_admin(self):
        self.context.session.query = mock.MagicMock()
        filter_mock = self.context.session.query.return_value

        db_api.ip_address_find(self.context.elevated(), address_type="foo")
        # NOTE(thomasem): Creates sqlalchemy.sql.elements.BinaryExpression
        # when using SQLAlchemy models in expressions.
        expected_filter = models.IPAddress.address_type == "foo"
        self.assertEqual(len(filter_mock.filter.call_args[0]), 1)
        # NOTE(thomasem): Unfortunately BinaryExpression.compare isn't
        # showing to be a reliable comparison, so using the string
        # representation which dumps the associated SQL for the filter.
        self.assertEqual(str(expected_filter), str(
            filter_mock.filter.call_args[0][0]))

    def test_ip_address_find_port_id_as_admin(self):
        self.context.session.query = mock.MagicMock()
        final_query_mock = self.context.session.query.return_value

        db_api.ip_address_find(self.context.elevated(), port_id="foo")
        # NOTE(thomasem): Creates sqlalchemy.sql.elements.BinaryExpression
        # when using SQLAlchemy models in expressions.
        expected_filter = models.IPAddress.ports.any(models.Port.id == "foo")
        self.assertEqual(len(final_query_mock.filter.call_args[0]), 1)
        self.assertEqual(str(expected_filter), str(
            final_query_mock.filter.call_args[0][0]))

    def test_ip_address_find_ip_address_object(self):
        ip_address = netaddr.IPAddress("192.168.10.1")
        try:
            db_api.ip_address_find(self.context, ip_address=ip_address,
                                   scope=db_api.ONE)
        except Exception as e:
            self.fail("Expected no exceptions: %s" % e)

    def test_ip_address_find_ip_address_list(self):
        ip_address = netaddr.IPAddress("192.168.10.1")
        try:
            db_api.ip_address_find(self.context, ip_address=[ip_address],
                                   scope=db_api.ONE)
        except Exception as e:
            self.fail("Expected no exceptions: %s" % e)

    def test_model_query_with_IPAddress(self):
        # NOTE: tenant_id filter will always be added
        test_model = models.IPAddress
        good_filter = {"network_id": [2]}
        result = db_api._model_query(self.context, test_model, good_filter)
        self.assertEqual(len(result), 2)
        bad_filter = {"ethertype": "IPv4"}
        result = db_api._model_query(self.context, test_model, bad_filter)
        self.assertEqual(len(result), 1)

    def test_model_query_with_MacAddress(self):
        test_model = models.MacAddress
        good_filter = {"deallocated": True}
        result = db_api._model_query(self.context, test_model, good_filter)
        self.assertEqual(len(result), 2)
        bad_filter = {"protocol": "ICMP"}
        result = db_api._model_query(self.context, test_model, bad_filter)
        self.assertEqual(len(result), 1)

    def test_model_query_with_Network(self):
        test_model = models.Network
        good_filter = {"name": ["BOB"]}
        result = db_api._model_query(self.context, test_model, good_filter)
        self.assertEqual(len(result), 2)
        bad_filter = {"deallocated": True}
        result = db_api._model_query(self.context, test_model, bad_filter)
        self.assertEqual(len(result), 1)

    def test_model_query_with_Port(self):
        test_model = models.Port
        good_filter = {"device_id": [123]}
        result = db_api._model_query(self.context, test_model, good_filter)
        self.assertEqual(len(result), 2)
        bad_filter = {"not_real": "BANANAS"}
        result = db_api._model_query(self.context, test_model, bad_filter)
        self.assertEqual(len(result), 1)

    def test_model_query_with_SecurityGroup(self):
        test_model = models.SecurityGroup
        good_filter = {"name": ["Abraham Lincoln"]}
        result = db_api._model_query(self.context, test_model, good_filter)
        self.assertEqual(len(result), 2)
        bad_filter = {"segment_id": [123]}
        result = db_api._model_query(self.context, test_model, bad_filter)
        self.assertEqual(len(result), 1)

    def test_model_query_with_SecurityGroupRule(self):
        test_model = models.SecurityGroupRule
        good_filter = {"ethertype": ["IPv4"]}
        result = db_api._model_query(self.context, test_model, good_filter)
        self.assertEqual(len(result), 2)
        bad_filter = {"made_up": "Moon Landing"}
        result = db_api._model_query(self.context, test_model, bad_filter)
        self.assertEqual(len(result), 1)

    def test_model_query_with_Subnet(self):
        test_model = models.Subnet
        good_filter = {"network_id": [42]}
        result = db_api._model_query(self.context, test_model, good_filter)
        self.assertEqual(len(result), 2)
        bad_filter = {"subnet_id": [123]}
        result = db_api._model_query(self.context, test_model, bad_filter)
        self.assertEqual(len(result), 1)

    def test_port_associate_ip(self):
        self.context.session.add = mock.Mock()
        mock_ports = [models.Port(id=str(x), network_id="2", ip_addresses=[])
                      for x in xrange(4)]
        mock_address = models.IPAddress(id="1", address=3232235876,
                                        address_readable="192.168.1.100",
                                        subnet_id="1", network_id="2",
                                        version=4, used_by_tenant_id="1")
        r = db_api.port_associate_ip(self.context, mock_ports, mock_address)
        self.assertEqual(len(r.associations), len(mock_ports))
        for assoc, port in zip(r.associations, mock_ports):
            self.assertEqual(assoc.port, port)
            self.assertEqual(assoc.ip_address, mock_address)
            self.assertEqual(assoc.enabled, False)

    def test_port_associate_ip_enable_port(self):
        self.context.session.add = mock.Mock()
        mock_port = models.Port(id="1", network_id="2", ip_addresses=[])
        mock_address = models.IPAddress(id="1", address=3232235876,
                                        address_readable="192.168.1.100",
                                        subnet_id="1", network_id="2",
                                        version=4, used_by_tenant_id="1")
        r = db_api.port_associate_ip(self.context, [mock_port], mock_address,
                                     enable_port="1")
        self.assertEqual(len(r.associations), 1)
        assoc = r.associations[0]
        self.assertEqual(assoc.port, mock_port)
        self.assertEqual(assoc.ip_address, mock_address)
        self.assertEqual(assoc.enabled, True)
        self.context.session.add.assert_called_once_with(assoc)

    def test_port_disassociate_ip(self):
        self.context.session.add = mock.Mock()
        self.context.session.delete = mock.Mock()
        mock_ports = [models.Port(id=str(x), network_id="2", ip_addresses=[])
                      for x in xrange(4)]
        mock_address = models.IPAddress(id="1", address=3232235876,
                                        address_readable="192.168.1.100",
                                        subnet_id="1", network_id="2",
                                        version=4, used_by_tenant_id="1")
        mock_assocs = []
        for p in mock_ports:
            assoc = models.PortIpAssociation()
            assoc.port_id = p.id
            assoc.port = p
            assoc.ip_address_id = mock_address.id
            assoc.ip_address = mock_address
            mock_assocs.append(assoc)

        r = db_api.port_disassociate_ip(self.context, mock_ports[1:3],
                                        mock_address)

        self.assertEqual(len(r.associations), 2)
        self.assertEqual(r.associations[0], mock_assocs[0])
        self.assertEqual(r.associations[1], mock_assocs[3])
        self.context.session.add.assert_called_once_with(r)
        self.context.session.delete.assert_has_calls(
            [mock.call(mock_assocs[1]), mock.call(mock_assocs[2])])

    @mock.patch("quark.db.api.get_ports_for_address")
    @mock.patch("quark.db.api.port_disassociate_ip")
    @mock.patch("quark.db.api.port_associate_ip")
    def test_update_port_associations_for_ip(self, associate_mock,
                                             disassociate_mock,
                                             get_associations_mock):
        self.context.session.add = mock.Mock()
        self.context.session.delete = mock.Mock()
        mock_ports = [models.Port(id=str(x), network_id="2", ip_addresses=[])
                      for x in xrange(4)]
        mock_address = models.IPAddress(id="1", address=3232235876,
                                        address_readable="192.168.1.100",
                                        subnet_id="1", network_id="2",
                                        version=4, used_by_tenant_id="1")
        mock_address.ports = mock_ports
        new_port_list = mock_ports[1:3]
        new_port_list.append(models.Port(id="4", network_id="2",
                             ip_addresses=[]))
        get_associations_mock.return_value = mock_ports
        # NOTE(thomasem): Should be the new address after associating
        # any new ports in the list.
        mock_new_address = associate_mock.return_value

        db_api.update_port_associations_for_ip(self.context,
                                               new_port_list,
                                               mock_address)

        associate_mock.assert_called_once_with(self.context,
                                               set([new_port_list[2]]),
                                               mock_address)

        disassociate_mock.assert_called_once_with(self.context,
                                                  set([mock_ports[0],
                                                       mock_ports[3]]),
                                                  mock_new_address)

    def test_update_port_sets_vlan_tag(self):
        self.context.session.add = mock.Mock()
        mock_port = models.Port(id=1, network_id="2", ip_addresses=[], tags=[])
        db_api.port_update(self.context, mock_port, vlan_id=1)
        self.assertEqual(mock_port.tags, [tags.VlanTag().serialize(1)])

    def test_create_port_sets_vlan_tag(self):
        self.context.session.add = mock.Mock()
        port_req = {"id": 1, "network_id": "2", "vlan_id": 1}
        new_port = db_api.port_create(self.context, **port_req)
        self.assertEqual(new_port.tags, [tags.VlanTag().serialize(1)])



"""Add size to quark_ip_policy

Revision ID: 3d22de205729
Revises: 2748e48cee3a
Create Date: 2014-07-04 23:53:09.531715

"""

# revision identifiers, used by Alembic.
revision = '3d22de205729'
down_revision = '3ed0c5a067f1'

from alembic import op
import sqlalchemy as sa

from quark.db.custom_types import INET


def upgrade():
    op.add_column('quark_ip_policy', sa.Column('size', INET()))


def downgrade():
    op.drop_column('quark_ip_policy', 'size')



# Copyright 2016 Rackspace Hosting Inc
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for# the specific language governing permissions and limitations
#  under the License.
from oslo_log import log as logging

from neutron_lib import exceptions as n_exc

from quark import exceptions as q_exc
from quark.plugin_modules import jobs as job_api
from quark.tests.functional.base import BaseFunctionalTest

LOG = logging.getLogger(__name__)


class QuarkJobs(BaseFunctionalTest):
    def setUp(self):
        super(QuarkJobs, self).setUp()
        self.action = "test action"
        self.tenant_id = "test_tenant"
        self.tenant_id2 = "test_tenant2"

    def test_create_job(self):
        job_body = dict(tenant_id=self.tenant_id, action=self.action,
                        completed=False)
        job_body = dict(job=job_body)
        job = job_api.create_job(self.admin_context, job_body)
        self.assertIsNotNone(job)
        self.assertFalse(job['completed'])
        self.assertEqual(self.tenant_id, job['tenant_id'])
        self.assertEqual(self.action, job['action'])

    def test_create_job_fail_non_admin(self):
        job_body = dict(tenant_id=self.tenant_id, action=self.action,
                        completed=False)
        job_body = dict(job=job_body)
        with self.assertRaises(n_exc.NotAuthorized):
            job_api.create_job(self.context, job_body)

    def test_get_jobs(self):
        job_body = dict(tenant_id=self.tenant_id, action=self.action,
                        completed=False)
        job_body = dict(job=job_body)
        job1 = job_api.create_job(self.admin_context, job_body)
        self.assertIsNotNone(job1)

        job_body = dict(tenant_id=self.tenant_id2, action=self.action,
                        completed=True)
        job_body = dict(job=job_body)
        job2 = job_api.create_job(self.admin_context, job_body)
        self.assertIsNotNone(job2)

        jobs = job_api.get_job(self.admin_context, job1['id'])
        self.assertFalse(type(jobs) in [list, tuple])
        job = jobs
        self.assertFalse(job['completed'])
        self.assertEqual(self.tenant_id, job['tenant_id'])
        self.assertEqual(self.action, job['action'])

        job = job_api.get_job(self.admin_context, job2['id'])
        self.assertTrue(job['completed'])
        self.assertEqual(self.tenant_id2, job['tenant_id'])
        self.assertEqual(self.action, job['action'])

        with self.assertRaises(q_exc.JobNotFound):
            job_api.get_job(self.admin_context, 'derp')

        jobs = job_api.get_jobs(self.admin_context)
        self.assertTrue(type(jobs) in [list, tuple])
        self.assertEqual(2, len(jobs))

        jobs = job_api.get_jobs(self.admin_context, completed=True)
        self.assertTrue(type(jobs) in [list, tuple])
        self.assertEqual(1, len(jobs))

        jobs = job_api.get_jobs(self.admin_context, completed=False)
        self.assertTrue(type(jobs) in [list, tuple])
        self.assertEqual(1, len(jobs))

        jobs = job_api.get_jobs(self.admin_context, completed='hello')
        self.assertTrue(type(jobs) in [list, tuple])
        self.assertEqual(0, len(jobs))

        jobs = job_api.get_jobs(self.admin_context, tenant_id=self.tenant_id)
        self.assertTrue(type(jobs) in [list, tuple])
        self.assertEqual(1, len(jobs))

        jobs = job_api.get_jobs(self.admin_context, tenant_id='derp')
        self.assertTrue(type(jobs) in [list, tuple])
        self.assertEqual(0, len(jobs))

    def test_get_job_different_non_admin(self):
        job_body = dict(tenant_id=self.context.tenant_id, action=self.action,
                        completed=False)
        job_body = dict(job=job_body)
        job1 = job_api.create_job(self.admin_context, job_body)
        self.assertIsNotNone(job1)

        job_body = dict(tenant_id=self.tenant_id2, action=self.action,
                        completed=True)
        job_body = dict(job=job_body)
        job2 = job_api.create_job(self.admin_context, job_body)
        self.assertIsNotNone(job2)

        jobs = job_api.get_jobs(self.context)
        self.assertTrue(type(jobs) in [list, tuple])

        self.assertEqual(1, len(jobs))
        self.assertEqual(self.context.tenant_id, jobs[0]['tenant_id'])

    def test_update_jobs(self):
        update_body = dict(completed=True)
        update_body = dict(job=update_body)

        with self.assertRaises(q_exc.JobNotFound):
            job_api.update_job(self.admin_context, 'derp', update_body)

        job_body = dict(tenant_id=self.tenant_id, action=self.action,
                        completed=False)
        job_body = dict(job=job_body)
        job1 = job_api.create_job(self.admin_context, job_body)
        self.assertIsNotNone(job1)

        job = job_api.get_job(self.admin_context, job1['id'])
        self.assertFalse(job['completed'])

        updated_job = job_api.update_job(self.admin_context, job1['id'],
                                         update_body)
        self.assertTrue(updated_job['completed'])

        job = job_api.get_job(self.admin_context, job1['id'])
        self.assertTrue(job['completed'])

    def test_update_job_fail_non_admin(self):
        update_body = dict(completed=True)
        update_body = dict(job=update_body)

        job_body = dict(tenant_id=self.tenant_id, action=self.action,
                        completed=False)
        job_body = dict(job=job_body)
        job1 = job_api.create_job(self.admin_context, job_body)
        self.assertIsNotNone(job1)

        job = job_api.get_job(self.admin_context, job1['id'])
        self.assertFalse(job['completed'])

        with self.assertRaises(n_exc.NotAuthorized):
            job_api.update_job(self.context, job1['id'], update_body)

        updated_job = job_api.update_job(self.admin_context, job1['id'],
                                         update_body)
        self.assertTrue(updated_job['completed'])

        job = job_api.get_job(self.admin_context, job1['id'])
        self.assertTrue(job['completed'])

    def test_delete_jobs(self):
        with self.assertRaises(q_exc.JobNotFound):
            job_api.delete_job(self.admin_context, 'derp')

        job_body = dict(tenant_id=self.tenant_id, action=self.action,
                        completed=False)
        job_body = dict(job=job_body)
        job1 = job_api.create_job(self.admin_context, job_body)
        self.assertIsNotNone(job1)

        job = job_api.get_job(self.admin_context, job1['id'])
        self.assertFalse(job['completed'])

        job_api.delete_job(self.admin_context, job1['id'])

        with self.assertRaises(q_exc.JobNotFound):
            job_api.get_job(self.admin_context, job1['id'])

        with self.assertRaises(q_exc.JobNotFound):
            job_api.delete_job(self.admin_context, job1['id'])

    def test_delete_job_fail_non_admin(self):
        with self.assertRaises(n_exc.NotAuthorized):
            job_api.delete_job(self.context, 'derp')

        job_body = dict(tenant_id=self.tenant_id, action=self.action,
                        completed=False)
        job_body = dict(job=job_body)
        job1 = job_api.create_job(self.admin_context, job_body)
        self.assertIsNotNone(job1)

        job = job_api.get_job(self.admin_context, job1['id'])
        self.assertFalse(job['completed'])

        with self.assertRaises(n_exc.NotAuthorized):
            job_api.delete_job(self.context, job1['id'])

        job_api.delete_job(self.admin_context, job1['id'])

        with self.assertRaises(q_exc.JobNotFound):
            job_api.get_job(self.context, job1['id'])



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json

from oslo_config import cfg

from quark import network_strategy
from quark.tests import test_base


class TestJSONStrategy(test_base.TestBase):
    def setUp(self):
        self.strategy = {"public_network": {"bridge": "xenbr0",
                                            "subnets": {"4": "public_v4",
                                                        "6": "public_v6"}}}
        strategy_json = json.dumps(self.strategy)
        cfg.CONF.set_override("default_net_strategy", strategy_json, "QUARK")

    def test_get_network(self):
        json_strategy = network_strategy.JSONStrategy()
        net = json_strategy.get_network("public_network")
        self.assertEqual(net["bridge"], "xenbr0")

    def test_split_network_ids(self):
        json_strategy = network_strategy.JSONStrategy()
        net_ids = ["foo_net", "public_network"]
        tenant, assignable = json_strategy.split_network_ids(net_ids)
        self.assertTrue("foo_net" in tenant)
        self.assertTrue("foo_net" not in assignable)
        self.assertTrue("public_network" not in tenant)
        self.assertTrue("public_network" in assignable)

    def test_split_subnet_ids(self):
        json_strategy = network_strategy.JSONStrategy()
        subnet_ids = ["tenant_subnet", "public_v6"]
        tenant, assignable = json_strategy.split_subnet_ids(subnet_ids)
        self.assertTrue("tenant_subnet" in tenant)
        self.assertTrue("tenant_subnet" not in assignable)
        self.assertTrue("public_v6" not in tenant)
        self.assertTrue("public_v6" in assignable)

    def test_is_provider_network(self):
        json_strategy = network_strategy.JSONStrategy()
        self.assertTrue(json_strategy.is_provider_network("public_network"))

    def test_is_not_provider_network(self):
        json_strategy = network_strategy.JSONStrategy()
        self.assertFalse(json_strategy.is_provider_network("tenant_network"))

    def test_is_provider_subnet(self):
        json_strategy = network_strategy.JSONStrategy()
        self.assertTrue(json_strategy.is_provider_subnet("public_v4"))

    def test_is_not_provider_subnet(self):
        json_strategy = network_strategy.JSONStrategy()
        self.assertFalse(json_strategy.is_provider_network("tenant_v4"))

    def test_get_provider_networks(self):
        json_strategy = network_strategy.JSONStrategy()
        expected = "public_network"
        nets = json_strategy.get_provider_networks()
        self.assertTrue(expected in nets)
        self.assertEqual(1, len(nets))

    def test_get_provider_subnets(self):
        json_strategy = network_strategy.JSONStrategy()
        expected = ["public_v4", "public_v6"]
        subs = json_strategy.get_provider_subnets()
        for sub in expected:
            self.assertTrue(sub in subs)
        self.assertEqual(2, len(subs))

    def test_get_network_for_subnet(self):
        json_strategy = network_strategy.JSONStrategy()
        net = json_strategy.get_network_for_subnet("public_v4")
        self.assertEqual("public_network", net)

    def test_get_network_for_subnet_matches_none(self):
        json_strategy = network_strategy.JSONStrategy()
        net = json_strategy.get_network_for_subnet("tenant_v4")
        self.assertIsNone(net)

    def test_subnet_ids_for_network(self):
        json_strategy = network_strategy.JSONStrategy()
        expected = ["public_v4", "public_v6"]
        subs = json_strategy.subnet_ids_for_network("public_network")
        for sub in expected:
            self.assertTrue(sub in subs)
        self.assertEqual(2, len(subs))

    def test_subnet_ids_for_network_matches_none(self):
        json_strategy = network_strategy.JSONStrategy()
        subs = json_strategy.subnet_ids_for_network("tenant_network")
        self.assertIsNone(subs)

    def test_get_provider_subnet_id(self):
        json_strategy = network_strategy.JSONStrategy()
        net_id = "public_network"
        ip_version = 4
        sub = json_strategy.get_provider_subnet_id(net_id, ip_version)
        self.assertEqual(sub, "public_v4")

    def test_get_provider_subnet_id_matches_none(self):
        json_strategy = network_strategy.JSONStrategy()
        net_id = "tenant_network"
        ip_version = 4
        sub = json_strategy.get_provider_subnet_id(net_id, ip_version)
        self.assertIsNone(sub)



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions as n_exc
from oslo_log import log as logging

from quark.db import api as db_api
from quark import exceptions as q_exc
from quark import plugin_views as v

from quark import segment_allocations

SA_REGISTRY = segment_allocations.REGISTRY

LOG = logging.getLogger(__name__)


def get_segment_allocation_range(context, id, fields=None):
    LOG.info("get_segment_allocation_range %s for tenant %s fields %s" %
             (id, context.tenant_id, fields))

    if not context.is_admin:
        raise n_exc.NotAuthorized()

    sa_range = db_api.segment_allocation_range_find(
        context, id=id, scope=db_api.ONE)

    if not sa_range:
        raise q_exc.SegmentAllocationRangeNotFound(
            segment_allocation_range_id=id)

    # Count up allocations so we can calculate how many are free.
    allocs = db_api.segment_allocation_find(
        context,
        segment_allocation_range_id=sa_range["id"],
        deallocated=False).count()
    return v._make_segment_allocation_range_dict(
        sa_range, allocations=allocs)


def get_segment_allocation_ranges(context, **filters):
    LOG.info("get_segment_allocation_ranges for tenant %s" % context.tenant_id)
    if not context.is_admin:
        raise n_exc.NotAuthorized()

    sa_ranges = db_api.segment_allocation_range_find(
        context, scope=db_api.ALL, **filters)
    return [v._make_segment_allocation_range_dict(m) for m in sa_ranges]


def create_segment_allocation_range(context, sa_range):
    LOG.info("create_segment_allocation_range for tenant %s"
             % context.tenant_id)
    if not context.is_admin:
        raise n_exc.NotAuthorized()

    sa_range = sa_range.get("segment_allocation_range")
    if not sa_range:
        raise n_exc.BadRequest(resource="segment_allocation_range",
                               msg=("segment_allocation_range not in "
                                    "request body."))

    # TODO(morgabra) Figure out how to get the api extension to validate this
    # for us.
    # parse required fields
    for k in ["first_id", "last_id", "segment_id", "segment_type"]:
        sa_range[k] = sa_range.get(k, None)
        if sa_range[k] is None:
            raise n_exc.BadRequest(
                resource="segment_allocation_range",
                msg=("Missing required key %s in request body." % (k)))

    # parse optional fields
    for k in ["do_not_use"]:
        sa_range[k] = sa_range.get(k, None)

    # use the segment registry to validate and create/populate the range
    if not SA_REGISTRY.is_valid_strategy(sa_range["segment_type"]):
        raise n_exc.BadRequest(
            resource="segment_allocation_range",
            msg=("Unknown segment type '%s'" % (k)))
    strategy = SA_REGISTRY.get_strategy(sa_range["segment_type"])

    # Create the new range
    with context.session.begin():
        new_range = strategy.create_range(context, sa_range)

    # Bulk-populate the range, this could take a while for huge ranges
    # (millions) so we do this in chunks outside the transaction. That
    # means we need to rollback the range creation if it fails for
    # whatever reason (and it will cascade delete any added allocations)
    try:
        strategy.populate_range(context, new_range)
    except Exception:
        LOG.exception("Failed to populate segment allocation range.")
        delete_segment_allocation_range(context, new_range["id"])
        raise

    return v._make_segment_allocation_range_dict(new_range)


def _delete_segment_allocation_range(context, sa_range):

    allocs = db_api.segment_allocation_find(
        context,
        segment_allocation_range_id=sa_range["id"],
        deallocated=False).count()

    if allocs:
        raise q_exc.SegmentAllocationRangeInUse(
            segment_allocation_range_id=sa_range["id"])
    db_api.segment_allocation_range_delete(context, sa_range)


def delete_segment_allocation_range(context, sa_id):
    """Delete a segment_allocation_range.

    : param context: neutron api request context
    : param id: UUID representing the segment_allocation_range to delete.
    """
    LOG.info("delete_segment_allocation_range %s for tenant %s" %
             (sa_id, context.tenant_id))
    if not context.is_admin:
        raise n_exc.NotAuthorized()

    with context.session.begin():
        sa_range = db_api.segment_allocation_range_find(
            context, id=sa_id, scope=db_api.ONE)
        if not sa_range:
            raise q_exc.SegmentAllocationRangeNotFound(
                segment_allocation_range_id=sa_id)
        _delete_segment_allocation_range(context, sa_range)



"""Added quark_mac_addresses index and unique constraint

Revision ID: 4358d1b8cc75
Revises: 56326c29d553
Create Date: 2014-06-04 15:45:23.481298

"""

# revision identifiers, used by Alembic.
revision = '4358d1b8cc75'
down_revision = '1817eef6373c'

from alembic import op


def upgrade():
    op.create_index(op.f('ix_quark_mac_addresses_deallocated'),
                    'quark_mac_addresses',
                    ['deallocated'],
                    unique=False)


def downgrade():
    op.drop_index(op.f('ix_quark_mac_addresses_deallocated'),
                  table_name='quark_mac_addresses')



#!/usr/bin/python
# Copyright 2015 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import contextlib

import mock

from quark.cache import security_groups_client
from quark.db import models
from quark import exceptions as q_exc
from quark.tests import test_base
from quark.tools import redis_sg_tool

TOOL_MOD = "quark.tools.redis_sg_tool.QuarkRedisTool"


def sg_client(args=None):
    args = args or {}
    return redis_sg_tool.QuarkRedisTool(args)


class QuarkRedisSgToolBase(test_base.TestBase):
    def setUp(self):
        super(QuarkRedisSgToolBase, self).setUp()
        neutron_cfg_patch = mock.patch("neutron.common.config.init")
        oslo_cfg_patch = mock.patch("oslo_config.cfg.CONF")
        neutron_cfg_patch.start()
        oslo_cfg_patch.start()
        self.addCleanup(neutron_cfg_patch.stop)
        self.addCleanup(oslo_cfg_patch.stop)


class QuarkRedisSgToolTestDispatch(QuarkRedisSgToolBase):
    def _client_dispatch(self, command):
        args = {"<command>": command}
        sg_client(args).dispatch()

    @mock.patch("%s.test_connection" % TOOL_MOD)
    def test_dispatch_test_connection(self, test_conn):
        self._client_dispatch("test-connection")
        test_conn.assert_called_with()

    @mock.patch("%s.vif_count" % TOOL_MOD)
    def test_dispatch_vifs_in_redis(self, vif_count):
        self._client_dispatch("vifs-in-redis")
        vif_count.assert_called_with()

    @mock.patch("%s.num_groups" % TOOL_MOD)
    def test_dispatch_num_groups(self, num_groups):
        self._client_dispatch("num-groups")
        num_groups.assert_called_with()

    @mock.patch("%s.ports_with_groups" % TOOL_MOD)
    def test_dispatch_ports_with_groups(self, ports_with_groups):
        self._client_dispatch("ports-with-groups")
        ports_with_groups.assert_called_with()

    @mock.patch("%s.purge_orphans" % TOOL_MOD)
    def test_dispatch_purge_orphans(self, purge_orphans):
        self._client_dispatch("purge-orphans")
        purge_orphans.assert_called_with(True)

    @mock.patch("%s.write_groups" % TOOL_MOD)
    def test_dispatch_write_groups(self, write_groups):
        self._client_dispatch("write-groups")
        write_groups.assert_called_with(True)

    @mock.patch("%s.test_connection" % TOOL_MOD)
    @mock.patch("%s.vif_count" % TOOL_MOD)
    @mock.patch("%s.num_groups" % TOOL_MOD)
    @mock.patch("%s.ports_with_groups" % TOOL_MOD)
    @mock.patch("%s.purge_orphans" % TOOL_MOD)
    @mock.patch("%s.write_groups" % TOOL_MOD)
    def test_dispatch_nothing(self, write_groups, purge_orphans,
                              ports_with_groups, num_groups, vif_count,
                              test_conn):
        self._client_dispatch("Nothing")
        self.assertFalse(write_groups.called)
        self.assertFalse(purge_orphans.called)
        self.assertFalse(ports_with_groups.called)
        self.assertFalse(num_groups.called)
        self.assertFalse(vif_count.called)
        self.assertFalse(test_conn.called)


class QuarkRedisSgToolTestConnection(QuarkRedisSgToolBase):
    @mock.patch("%s._get_connection" % TOOL_MOD)
    def test_connected(self, get_conn):
        get_conn.return_value = True
        cli = sg_client()
        self.assertTrue(cli.test_connection())

    @mock.patch("%s._get_connection" % TOOL_MOD)
    def test_not_connected(self, get_conn):
        get_conn.return_value = False
        cli = sg_client()
        self.assertFalse(cli.test_connection())


class QuarkRedisSgToolVifCount(QuarkRedisSgToolBase):
    @mock.patch("%s._get_connection" % TOOL_MOD)
    def test_vif_count(self, get_conn):
        conn_mock = mock.MagicMock()
        get_conn.return_value = conn_mock
        cli = sg_client()
        cli.vif_count()
        conn_mock.vif_keys.assert_called_with(
            field=security_groups_client.SECURITY_GROUP_HASH_ATTR)


class QuarkRedisSgToolNumGroups(QuarkRedisSgToolBase):
    @mock.patch("neutron.context.get_admin_context")
    @mock.patch("quark.db.api.security_group_count")
    def test_num_groups(self, group_count, get_admin_ctxt):
        ctxt_mock = mock.MagicMock()
        get_admin_ctxt.return_value = ctxt_mock

        cli = sg_client()
        cli.num_groups()
        group_count.assert_called_with(ctxt_mock)


class QuarkRedisSgToolPortsWithGroups(QuarkRedisSgToolBase):
    @mock.patch("neutron.context.get_admin_context")
    @mock.patch("quark.db.api.ports_with_security_groups_count")
    def test_num_groups(self, db_ports_groups, get_admin_ctxt):
        db_ports_groups.return_value = mock.MagicMock()
        ctxt_mock = mock.MagicMock()
        get_admin_ctxt.return_value = ctxt_mock

        cli = sg_client()
        cli.ports_with_groups()
        db_ports_groups.assert_called_with(ctxt_mock)


class QuarkRedisSgToolPurgeOrphans(QuarkRedisSgToolBase):
    @contextlib.contextmanager
    def _stubs(self):
        ports = [{"device_id": 1, "mac_address": 1}]
        vifs = ["1.1", "2.2", "3.3"]
        with contextlib.nested(
            mock.patch("neutron.context.get_admin_context"),
            mock.patch("quark.db.api.ports_with_security_groups_find"),
            mock.patch("%s._get_connection" % TOOL_MOD)
        ) as (get_admin_ctxt, db_ports_groups, get_conn):
            connection_mock = mock.MagicMock()
            get_conn.return_value = connection_mock
            ports_with_groups_mock = mock.MagicMock()
            db_ports_groups.return_value = ports_with_groups_mock
            ctxt_mock = mock.MagicMock()
            get_admin_ctxt.return_value = ctxt_mock
            ports_with_groups_mock.all.return_value = ports
            connection_mock.vif_keys.return_value = vifs
            yield get_conn, connection_mock, db_ports_groups, ctxt_mock

    def test_purge_orphans_dry_run(self):
        with self._stubs() as (get_conn, connection_mock, db_ports_groups,
                               ctxt_mock):
            cli = sg_client()
            cli.purge_orphans(dryrun=True)

            connection_mock.vif_key.assert_any_call(1, 1)
            db_ports_groups.assert_called_with(ctxt_mock)
            connection_mock.delete_key.assert_not_called()
            self.assertTrue(get_conn.call_count, 1)

    def test_purge_orphans(self):
        with self._stubs() as (get_conn, connection_mock, db_ports_groups,
                               ctxt_mock):
            cli = sg_client()
            cli.purge_orphans(dryrun=False)

            db_ports_groups.assert_called_with(ctxt_mock)
            connection_mock.vif_key.assert_any_call(1, 1)
            connection_mock.delete_key.assert_any_call("2.2")
            connection_mock.delete_key.assert_any_call("3.3")

    @mock.patch("time.sleep")
    def test_purge_orphans_raises(self, sleep):
        retry_delay = 1
        retries = 1
        with self._stubs() as (get_conn, connection_mock, db_ports_groups,
                               ctxt_mock):
            redis_exc = q_exc.RedisConnectionFailure
            connection_mock.delete_key.side_effect = redis_exc
            cli = sg_client({"--retry-delay": retry_delay,
                             "--retries": retries})
            cli.purge_orphans(dryrun=False)

            db_ports_groups.assert_called_with(ctxt_mock)
            get_conn.assert_called_with(giveup=False)
            connection_mock.vif_key.assert_any_call(1, 1)
            connection_mock.delete_key.assert_any_call("2.2")
            connection_mock.delete_key.assert_any_call("3.3")
            sleep.assert_called_with(1)


class QuarkRedisSgToolWriteGroups(QuarkRedisSgToolBase):
    @contextlib.contextmanager
    def _stubs(self):
        ports = [{"device_id": 1, "mac_address": 1}]
        vifs = ["1.1", "2.2", "3.3"]
        security_groups = [{"id": 1, "name": "test_group"}]

        with contextlib.nested(
            mock.patch("neutron.context.get_admin_context"),
            mock.patch("quark.db.api.security_group_rule_find"),
            mock.patch("quark.db.api.ports_with_security_groups_find"),
            mock.patch("%s._get_connection" % TOOL_MOD)
        ) as (get_admin_ctxt, rule_find, db_ports_groups, get_conn):
            connection_mock = mock.MagicMock()
            get_conn.return_value = connection_mock
            ports_with_groups_mock = mock.MagicMock()

            port_mods = []
            sg_mods = [models.SecurityGroup(**sg) for sg in security_groups]
            for port in ports:
                port_mod = models.Port(**port)
                port_mod.security_groups = sg_mods
                port_mods.append(port_mod)

            sg_rule = models.SecurityGroupRule()
            rule_find.return_value = [sg_rule]

            db_ports_groups.return_value = ports_with_groups_mock
            ctxt_mock = mock.MagicMock()
            get_admin_ctxt.return_value = ctxt_mock
            ports_with_groups_mock.all.return_value = port_mods
            connection_mock.vif_keys.return_value = vifs
            connection_mock.serialize_rules.return_value = "rules"
            yield (get_conn, connection_mock, db_ports_groups, ctxt_mock,
                   sg_rule)

    def test_write_groups_dryrun(self):
        with self._stubs() as (get_conn, connection_mock, db_ports_groups,
                               ctxt_mock, sg_rule):
            cli = sg_client()
            cli.write_groups(dryrun=True)
            connection_mock.vif_keys.assert_called_with()
            connection_mock.get_rules_for_port.assert_called_with(1, 1)

            self.assertTrue(get_conn.call_count, 1)

    def test_write_groups(self):
        with self._stubs() as (get_conn, connection_mock, db_ports_groups,
                               ctxt_mock, sg_rule):
            cli = sg_client()
            cli.write_groups(dryrun=False)
            self.assertFalse(connection_mock.vif_keys.called)
            self.assertFalse(connection_mock.get_rules_for_port.called)
            self.assertFalse(connection_mock.get_rules_for_port.called)
            connection_mock.serialize_rules.assert_called_with([sg_rule])
            connection_mock.apply_rules.assert_called_with(1, 1, "rules")

            self.assertTrue(get_conn.call_count, 1)

    @mock.patch("time.sleep")
    def test_write_groups_raises(self, sleep):
        with self._stubs() as (get_conn, connection_mock, db_ports_groups,
                               ctxt_mock, sg_rule):
            retry_delay = 1
            retries = 1
            cli = sg_client({"--retry-delay": retry_delay,
                             "--retries": retries})
            redis_exc = q_exc.RedisConnectionFailure
            connection_mock.apply_rules.side_effect = redis_exc

            cli.write_groups(dryrun=False)
            self.assertFalse(connection_mock.vif_keys.called)
            self.assertFalse(connection_mock.get_rules_for_port.called)
            self.assertFalse(connection_mock.get_rules_for_port.called)
            connection_mock.serialize_rules.assert_called_with([sg_rule])
            sleep.assert_called_with(1)
            self.assertTrue(get_conn.call_count, 2)
            get_conn.assert_any_call(giveup=False)



import sys

from neutron.common import config
from neutron.db import api as neutron_db_api
from oslo_config import cfg
from oslo_db.exception import DBDuplicateEntry
from oslo_log import log as logging

from quark.db import models

LOG = logging.getLogger(__name__)


def main():
    config.init(sys.argv[1:])
    if not cfg.CONF.config_file:
        sys.exit(_("ERROR: Unable to find configuration file via the default"
                   " search paths (~/.neutron/, ~/, /etc/neutron/, /etc/) and"
                   " the '--config-file' option!"))

    config.setup_logging()
    session = neutron_db_api.get_session()
    subnets = [
        models.Subnet(name="public_v4",
                      network_id="00000000-0000-0000-0000-000000000000",
                      id="00000000-0000-0000-0000-000000000000",
                      tenant_id="rackspace",
                      segment_id="rackspace",
                      do_not_use=True,
                      _cidr="0.0.0.0/0"),
        models.Subnet(name="public_v6",
                      network_id="00000000-0000-0000-0000-000000000000",
                      id="11111111-1111-1111-1111-111111111111",
                      tenant_id="rackspace",
                      segment_id="rackspace",
                      do_not_use=True,
                      _cidr="::/0"),
        models.Subnet(name="private_v4",
                      network_id="11111111-1111-1111-1111-111111111111",
                      id="22222222-2222-2222-2222-222222222222",
                      tenant_id="rackspace",
                      segment_id="rackspace",
                      do_not_use=True,
                      _cidr="0.0.0.0/0"),
        models.Subnet(name="private_v6",
                      network_id="11111111-1111-1111-1111-111111111111",
                      id="33333333-3333-3333-3333-333333333333",
                      tenant_id="rackspace",
                      segment_id="rackspace",
                      do_not_use=True,
                      _cidr="::/0")]
    try:
        session.bulk_save_objects(subnets)
    except DBDuplicateEntry:
        LOG.warn("Provider subnets previously inserted into database")


if __name__ == "__main__":
    main()



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#

from oslo_log import log as logging

LOG = logging.getLogger(__name__)


class BaseDriver(object):
    """Base interface for all Quark drivers.

    Usable as a replacement for the sample plugin.
    """
    def __init__(self):
        self.load_config()

    def load_config(self):
        LOG.info("load_config")

    @classmethod
    def get_name(klass):
        return "BASE"

    def get_connection(self):
        LOG.info("get_connection")

    def select_ipam_strategy(self, network_id, network_strategy, **kwargs):
        LOG.info("Selecting IPAM strategy for network_id:%s "
                 "network_strategy:%s" % (network_id, network_strategy))
        LOG.info("Selected IPAM strategy: %s" % (network_strategy))
        return network_strategy

    def create_network(self, context, network_name, tags=None,
                       network_id=None, **kwargs):
        LOG.info("create_network %s %s %s" % (context, network_name,
                                              tags))

    def delete_network(self, context, network_id, **kwargs):
        LOG.info("delete_network %s" % network_id)

    def diag_network(self, context, network_id, **kwargs):
        LOG.info("diag_network %s" % network_id)
        return {}

    def create_port(self, context, network_id, port_id, **kwargs):
        LOG.info("create_port %s %s %s" % (context.tenant_id, network_id,
                                           port_id))
        return {"uuid": port_id}

    def update_port(self, context, port_id, **kwargs):
        LOG.info("update_port %s %s" % (context.tenant_id, port_id))
        return {"uuid": port_id}

    def delete_port(self, context, port_id, **kwargs):
        LOG.info("delete_port %s %s" % (context.tenant_id, port_id))

    def diag_port(self, context, network_id, **kwargs):
        LOG.info("diag_port %s" % network_id)
        return {}

    def create_security_group(self, context, group_name, **group):
        LOG.info("Creating security profile %s for tenant %s" %
                 (group_name, context.tenant_id))

    def delete_security_group(self, context, group_id, **kwargs):
        LOG.info("Deleting security profile %s for tenant %s" %
                 (group_id, context.tenant_id))

    def update_security_group(self, context, group_id, **group):
        LOG.info("Updating security profile %s for tenant %s" %
                 (group_id, context.tenant_id))

    def create_security_group_rule(self, context, group_id, rule):
        LOG.info("Creating security rule on group %s for tenant %s" %
                 (group_id, context.tenant_id))

    def delete_security_group_rule(self, context, group_id, rule):
        LOG.info("Deleting security rule on group %s for tenant %s" %
                 (group_id, context.tenant_id))



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
#  under the License.

import contextlib

import datetime
import mock
import netaddr
from neutron.common import exceptions as ex

from quark.billing import IP_ASSOC
from quark.billing import IP_DISASSOC
from quark.db import models
from quark import exceptions as q_ex
from quark.plugin_modules import floating_ips
from quark.tests import test_quark_plugin


class TestRemoveFloatingIPs(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, flip=None):
        flip_model = None
        if flip:
            flip_model = models.IPAddress()
            flip_model.update(flip)

        with contextlib.nested(
            mock.patch("quark.db.api.floating_ip_find"),
            mock.patch("quark.db.api.floating_ip_disassociate_fixed_ip"),
            mock.patch("quark.db.api.port_disassociate_ip"),
            mock.patch("quark.db.api.ip_address_deallocate"),
            mock.patch("quark.ipam.QuarkIpam.deallocate_ip_address"),
            mock.patch("quark.drivers.unicorn_driver.UnicornDriver"
                       ".remove_floating_ip"),
            mock.patch("quark.billing.notify"),
            mock.patch("quark.billing.build_payload")
        ) as (flip_find, db_fixed_ip_disassoc, db_port_disassoc, db_dealloc,
              mock_dealloc, mock_remove_flip, notify, build_payload):
            flip_find.return_value = flip_model
            build_payload.return_value = {'respek': '4reelz'}
            yield

    def test_delete_floating_by_ip_address_id(self):
        flip = dict(id=1, address=3232235876, address_readable="192.168.1.100",
                    subnet_id=1, network_id=2, version=4, used_by_tenant_id=1,
                    network=dict(ipam_strategy="ANY"))
        with self._stubs(flip=flip):
            self.plugin.delete_floatingip(self.context, 1)

    def test_delete_floating_by_when_ip_address_does_not_exists_fails(self):
        with self._stubs():
            with self.assertRaises(q_ex.FloatingIpNotFound):
                self.plugin.delete_floatingip(self.context, 1)


class TestFloatingIPUtilityMethods(test_quark_plugin.TestQuarkPlugin):
    def test_get_next_available_fixed_ip_with_single_fixed_ip(self):
        port = models.Port()
        port.update(dict(id=1))

        fixed_ip_addr = netaddr.IPAddress('192.168.0.1')
        fixed_ip = models.IPAddress()
        fixed_ip.update(dict(address_type="fixed", address=int(fixed_ip_addr),
                             version=4, address_readable=str(fixed_ip_addr),
                             allocated_at=datetime.datetime.now()))

        port.ip_addresses.append(fixed_ip)

        next_fixed_ip = floating_ips._get_next_available_fixed_ip(port)

        self.assertEqual(next_fixed_ip["address_readable"], '192.168.0.1')

    def test_get_next_available_fixed_ip_with_mult_fixed_ips(self):
        port = models.Port()
        port.update(dict(id=1))

        for ip_addr in ["192.168.0.1", "192.168.0.2", "192.168.0.3"]:
            fixed_ip_addr = netaddr.IPAddress(ip_addr)
            fixed_ip = models.IPAddress()
            fixed_ip.update(dict(address_type="fixed",
                                 address=int(fixed_ip_addr),
                                 version=4,
                                 address_readable=str(fixed_ip_addr),
                                 allocated_at=datetime.datetime.now()))

            port.ip_addresses.append(fixed_ip)

        next_fixed_ip = floating_ips._get_next_available_fixed_ip(port)

        self.assertEqual(next_fixed_ip["address_readable"], '192.168.0.1')

    def test_get_next_available_fixed_ip_with_no_avail_fixed_ips(self):
        port = models.Port()
        port.update(dict(id=1))

        fixed_ip_addr = netaddr.IPAddress("192.168.0.1")
        fixed_ip = models.IPAddress()
        fixed_ip.update(dict(address_type="fixed",
                             address=int(fixed_ip_addr),
                             version=4,
                             address_readable=str(fixed_ip_addr),
                             allocated_at=datetime.datetime.now()))

        flip_addr = netaddr.IPAddress("10.0.0.1")
        flip = models.IPAddress()
        flip.update(dict(address_type="floating",
                         address=int(flip_addr),
                         version=4,
                         address_readable=str(flip_addr),
                         allocated_at=datetime.datetime.now()))
        flip.fixed_ip = fixed_ip

        port.ip_addresses.append(fixed_ip)
        port.ip_addresses.append(flip)

        fixed_ip_addr = netaddr.IPAddress("192.168.0.2")
        fixed_ip = models.IPAddress()
        fixed_ip.update(dict(address_type="fixed",
                             address=int(fixed_ip_addr),
                             version=4,
                             address_readable=str(fixed_ip_addr),
                             allocated_at=datetime.datetime.now()))

        flip_addr = netaddr.IPAddress("10.0.0.2")
        flip = models.IPAddress()
        flip.update(dict(address_type="floating",
                         address=int(flip_addr),
                         version=4,
                         address_readable=str(flip_addr),
                         allocated_at=datetime.datetime.now()))
        flip.fixed_ip = fixed_ip

        port.ip_addresses.append(fixed_ip)
        port.ip_addresses.append(flip)

        next_fixed_ip = floating_ips._get_next_available_fixed_ip(port)

        self.assertEqual(next_fixed_ip, None)

    def test_get_next_available_fixed_ip_with_avail_fixed_ips(self):
        port = models.Port()
        port.update(dict(id=1))

        fixed_ip_addr = netaddr.IPAddress("192.168.0.1")
        fixed_ip = models.IPAddress()
        fixed_ip.update(dict(address_type="fixed",
                             address=int(fixed_ip_addr),
                             version=4,
                             address_readable=str(fixed_ip_addr),
                             allocated_at=datetime.datetime.now()))

        flip_addr = netaddr.IPAddress("10.0.0.1")
        flip = models.IPAddress()
        flip.update(dict(address_type="floating",
                         address=int(flip_addr),
                         version=4,
                         address_readable=str(flip_addr),
                         allocated_at=datetime.datetime.now()))
        flip.fixed_ip = fixed_ip

        port.ip_addresses.append(fixed_ip)
        port.ip_addresses.append(flip)

        fixed_ip_addr = netaddr.IPAddress("192.168.0.2")
        fixed_ip = models.IPAddress()
        fixed_ip.update(dict(address_type="fixed",
                             address=int(fixed_ip_addr),
                             version=4,
                             address_readable=str(fixed_ip_addr),
                             allocated_at=datetime.datetime.now()))

        port.ip_addresses.append(fixed_ip)
        port.ip_addresses.append(flip)

        next_fixed_ip = floating_ips._get_next_available_fixed_ip(port)

        self.assertEqual(next_fixed_ip["address_readable"], "192.168.0.2")


class TestCreateFloatingIPs(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, flip=None, port=None, ips=None, network=None):
        port_model = None
        if port:
            port_model = models.Port()
            port_model.update(dict(port=port))
            if ips:
                for ip in ips:
                    ip_model = models.IPAddress()
                    ip_model.update(ip)
                    addr_type = ip.get("address_type")
                    if addr_type == "floating" and "fixed_ip_addr" in ip:
                        fixed_ip = models.IPAddress()
                        fixed_ip.update(next(ip_addr for ip_addr in ips
                                             if (ip_addr["address_readable"] ==
                                                 ip["fixed_ip_addr"])))
                        ip_model.fixed_ip = fixed_ip
                    port_model.ip_addresses.append(ip_model)

        flip_model = None
        if flip:
            flip_model = models.IPAddress()
            flip_model.update(flip)

        net_model = None
        if network:
            net_model = models.Network()
            net_model.update(network)

        def _alloc_ip(context, new_addr, net_id, port_m, *args, **kwargs):
            new_addr.append(flip_model)

        def _port_assoc(context, ports, addr, enable_port=None):
            addr.ports = ports
            return addr

        def _flip_fixed_ip_assoc(context, addr, fixed_ip):
            addr.fixed_ips.append(fixed_ip)
            return addr

        with contextlib.nested(
            mock.patch("quark.db.api.floating_ip_find"),
            mock.patch("quark.db.api.network_find"),
            mock.patch("quark.db.api.port_find"),
            mock.patch("quark.ipam.QuarkIpam.allocate_ip_address"),
            mock.patch("quark.drivers.unicorn_driver.UnicornDriver"
                       ".register_floating_ip"),
            mock.patch("quark.db.api.port_associate_ip"),
            mock.patch("quark.db.api.floating_ip_associate_fixed_ip"),
            mock.patch("quark.billing.notify"),
            mock.patch("quark.billing.build_payload")
        ) as (flip_find, net_find, port_find, alloc_ip, mock_reg_flip,
              port_assoc, fixed_ip_assoc, notify, build_payload):
            flip_find.return_value = flip_model
            net_find.return_value = net_model
            port_find.return_value = port_model
            alloc_ip.side_effect = _alloc_ip
            port_assoc.side_effect = _port_assoc
            fixed_ip_assoc.side_effect = _flip_fixed_ip_assoc
            build_payload.return_value = {}
            yield

    def test_create_with_a_port(self):
        floating_ip_addr = netaddr.IPAddress("10.0.0.1")
        floating_ip = dict(id=1, address=int(floating_ip_addr), version=4,
                           address_readable=str(floating_ip_addr), subnet_id=1,
                           network_id=2, used_by_tenant_id=1)
        network = dict(id="00000000-0000-0000-0000-000000000000",
                       ipam_strategy="ANY")

        fixed_ip_addr = netaddr.IPAddress("192.168.0.1")
        fixed_ips = [dict(address_type="fixed", address=int(fixed_ip_addr),
                          version=4, address_readable=str(fixed_ip_addr),
                          allocated_at=datetime.datetime.now())]
        port = dict(id="abcdefgh-1111-2222-3333-1234567890ab")

        with self._stubs(flip=floating_ip, port=port,
                         ips=fixed_ips, network=network):
            request = dict(floating_network_id=network["id"],
                           port_id=port["id"])
            flip = self.plugin.create_floatingip(self.context,
                                                 dict(floatingip=request))
            self.assertEqual(flip["floating_ip_address"], "10.0.0.1")
            self.assertEqual(flip["fixed_ip_address"], "192.168.0.1")

    def test_create_without_a_port(self):
        floating_ip_addr = netaddr.IPAddress("10.0.0.1")
        floating_ip = dict(id=1, address=int(floating_ip_addr), version=4,
                           address_readable=str(floating_ip_addr), subnet_id=1,
                           network_id=2, used_by_tenant_id=1)
        network = dict(id="00000000-0000-0000-0000-000000000000",
                       ipam_strategy="ANY")

        fixed_ip_addr = netaddr.IPAddress("192.168.0.1")
        fixed_ips = [dict(address_type="fixed", address=int(fixed_ip_addr),
                          version=4, address_readable=str(fixed_ip_addr),
                          allocated_at=datetime.datetime.now())]

        with self._stubs(flip=floating_ip, port=None,
                         ips=fixed_ips, network=network):
            request = dict(floating_network_id=network["id"], port_id=None)
            flip = self.plugin.create_floatingip(self.context,
                                                 dict(floatingip=request))
            self.assertEqual(flip["floating_ip_address"], "10.0.0.1")
            self.assertEqual(flip.get("fixed_ip_address"), None)

    def test_create_with_fixed_ip_specified(self):
        floating_ip_addr = netaddr.IPAddress("10.0.0.1")
        floating_ip = dict(id=1, address=int(floating_ip_addr), version=4,
                           address_readable=str(floating_ip_addr), subnet_id=1,
                           network_id=2, used_by_tenant_id=1)
        network = dict(id="00000000-0000-0000-0000-000000000000",
                       ipam_strategy="ANY")

        fixed_ips = []
        for ip_addr in ["192.168.0.1", "192.168.0.2"]:
            fixed_ip_addr = netaddr.IPAddress(ip_addr)
            fixed_ips.append(dict(address_type="fixed", version=4,
                                  address=int(fixed_ip_addr),
                                  address_readable=str(fixed_ip_addr),
                                  allocated_at=datetime.datetime.now()))

        port = dict(id="abcdefgh-1111-2222-3333-1234567890ab")

        with self._stubs(flip=floating_ip, port=port,
                         ips=fixed_ips, network=network):
            request = dict(floating_network_id=network["id"],
                           port_id=port["id"], fixed_ip_address="192.168.0.2")
            flip = self.plugin.create_floatingip(self.context,
                                                 dict(floatingip=request))
            self.assertEqual(flip["floating_ip_address"], "10.0.0.1")
            self.assertEqual(flip["fixed_ip_address"], "192.168.0.2")

    def test_create_with_floating_ip_specified(self):
        floating_ip_addr = netaddr.IPAddress("10.0.0.1")
        floating_ip = dict(id=1, address=int(floating_ip_addr), version=4,
                           address_readable=str(floating_ip_addr), subnet_id=1,
                           network_id=2, used_by_tenant_id=1)
        network = dict(id="00000000-0000-0000-0000-000000000000",
                       ipam_strategy="ANY")

        fixed_ip_addr = netaddr.IPAddress("192.168.0.1")
        fixed_ips = [dict(address_type="fixed", address=int(fixed_ip_addr),
                          version=4, address_readable=str(fixed_ip_addr),
                          allocated_at=datetime.datetime.now())]
        port = dict(id=2)

        with self._stubs(flip=floating_ip, port=port,
                         ips=fixed_ips, network=network):
            request = dict(floating_network_id=network["id"],
                           port_id=port["id"], floating_ip_address="10.0.0.1")
            flip = self.plugin.create_floatingip(self.context,
                                                 dict(floatingip=request))
            self.assertEqual(flip["floating_ip_address"], "10.0.0.1")
            self.assertEqual(flip["fixed_ip_address"], "192.168.0.1")

    def test_create_without_network_id_fails(self):
        with self._stubs():
            with self.assertRaises(ex.BadRequest):
                request = dict(port_id=2, floating_ip_address="10.0.0.1")
                self.plugin.create_floatingip(self.context,
                                              dict(floatingip=request))

    def test_create_with_invalid_network_fails(self):
        with self._stubs():
            with self.assertRaises(ex.NetworkNotFound):
                request = dict(floating_network_id=123,
                               port_id=2, floating_ip_address="10.0.0.1")
                self.plugin.create_floatingip(self.context,
                                              dict(floatingip=request))

    def test_create_with_invalid_port_fails(self):
        network = dict(id="00000000-0000-0000-0000-000000000000",
                       ipam_strategy="ANY")

        with self._stubs(network=network):
            with self.assertRaises(ex.PortNotFound):
                request = dict(floating_network_id=network["id"],
                               port_id=2, floating_ip_address="10.0.0.1")
                self.plugin.create_floatingip(self.context,
                                              dict(floatingip=request))

    def test_create_with_invalid_fixed_ip_for_port_fails(self):
        network = dict(id="00000000-0000-0000-0000-000000000000",
                       ipam_strategy="ANY")

        fixed_ip_addr = netaddr.IPAddress("192.168.0.1")
        fixed_ips = [dict(address_type="fixed", version=4,
                          address=int(fixed_ip_addr),
                          address_readable=str(fixed_ip_addr),
                          allocated_at=datetime.datetime.now())]

        port = dict(id="abcdefgh-1111-2222-3333-1234567890ab")

        with self._stubs(port=port, ips=fixed_ips, network=network):
            with self.assertRaises(
                    q_ex.FixedIpDoesNotExistsForPort):
                request = dict(floating_network_id=network["id"],
                               port_id=port["id"],
                               fixed_ip_address="192.168.0.2")
                flip = self.plugin.create_floatingip(self.context,
                                                     dict(floatingip=request))
                self.assertEqual(flip["address_readable"], "10.0.0.1")
                self.assertEqual(flip.fixed_ip["address_readable"],
                                 "192.168.0.2")

    def test_create_with_port_and_fixed_ip_with_existing_flip_fails(self):
        network = dict(id="00000000-0000-0000-0000-000000000000",
                       ipam_strategy="ANY")

        fixed_ip_addr = netaddr.IPAddress("192.168.0.1")
        fixed_ip = dict(address_type="fixed", version=4,
                        address=int(fixed_ip_addr),
                        address_readable=str(fixed_ip_addr),
                        allocated_at=datetime.datetime.now())

        floating_ip_addr = netaddr.IPAddress("10.0.0.1")
        floating_ip = dict(address_type="floating", version=4,
                           address=int(floating_ip_addr),
                           address_readable=str(floating_ip_addr),
                           allocated_at=datetime.datetime.now(),
                           fixed_ip_addr="192.168.0.1")

        ips = [fixed_ip, floating_ip]

        port = dict(id="abcdefgh-1111-2222-3333-1234567890ab")

        with self._stubs(port=port, ips=ips, network=network):
            with self.assertRaises(
                    q_ex.PortAlreadyContainsFloatingIp):
                request = dict(floating_network_id=network["id"],
                               port_id=port["id"],
                               fixed_ip_address="192.168.0.1")
                self.plugin.create_floatingip(self.context,
                                              dict(floatingip=request))

    def test_create_when_port_has_no_fixed_ips_fails(self):
        network = dict(id="00000000-0000-0000-0000-000000000000",
                       ipam_strategy="ANY")

        port = dict(id="abcdefgh-1111-2222-3333-1234567890ab")

        with self._stubs(port=port, network=network):
            with self.assertRaises(
                    q_ex.NoAvailableFixedIpsForPort):
                request = dict(floating_network_id=network["id"],
                               port_id=port["id"])
                self.plugin.create_floatingip(self.context,
                                              dict(floatingip=request))

    def test_create_when_port_has_no_available_fixed_ips_fails(self):
        network = dict(id="00000000-0000-0000-0000-000000000000",
                       ipam_strategy="ANY")

        fixed_ip_addr = netaddr.IPAddress("192.168.0.1")
        fixed_ip = dict(address_type="fixed", version=4,
                        address=int(fixed_ip_addr),
                        address_readable=str(fixed_ip_addr),
                        allocated_at=datetime.datetime.now())

        floating_ip_addr = netaddr.IPAddress("10.0.0.1")
        floating_ip = dict(address_type="floating", version=4,
                           address=int(floating_ip_addr),
                           address_readable=str(floating_ip_addr),
                           allocated_at=datetime.datetime.now(),
                           fixed_ip_addr="192.168.0.1")

        ips = [fixed_ip, floating_ip]

        port = dict(id="abcdefgh-1111-2222-3333-1234567890ab")

        with self._stubs(port=port, ips=ips, network=network):
            with self.assertRaises(
                    q_ex.NoAvailableFixedIpsForPort):
                request = dict(floating_network_id=network["id"],
                               port_id=port["id"])
                self.plugin.create_floatingip(self.context,
                                              dict(floatingip=request))


class TestUpdateFloatingIPs(test_quark_plugin.TestQuarkPlugin):

    def setUp(self):
        super(TestUpdateFloatingIPs, self).setUp()
        # NOTE(blogan): yuck yuck yuck, but since the models are being mocked
        # and not attached to the session, the refresh call will fail.
        old_refresh = self.context.session.refresh

        def reset_refresh(context):
            context.session.refresh = old_refresh

        self.context.session.refresh = mock.Mock()
        self.addCleanup(reset_refresh, self.context)

    @contextlib.contextmanager
    def _stubs(self, flip=None, curr_port=None, new_port=None, ips=None,
               subnet=None, network=None):
        curr_port_model = None
        if curr_port:
            curr_port_model = models.Port()
            curr_port_model.update(curr_port)

        new_port_model = None
        if new_port:
            new_port_model = models.Port()
            new_port_model.update(new_port)
            if ips:
                for ip in ips:
                    ip_model = models.IPAddress()
                    ip_model.update(ip)
                    addr_type = ip.get("address_type")
                    if addr_type == "floating" and "fixed_ip_addr" in ip:
                        fixed_ip = models.IPAddress()
                        fixed_ip.update(next(ip_addr for ip_addr in ips
                                             if (ip_addr["address_readable"] ==
                                                 ip["fixed_ip_addr"])))
                        ip_model.fixed_ip = fixed_ip
                    new_port_model.ip_addresses.append(ip_model)

        subnet_model = None
        if subnet:
            subnet_model = models.Subnet()
            subnet_model.update(subnet)
        else:
            subnet_model = models.Subnet(id=1, network_id=1, name=1,
                                         tenant_id=1, ip_version=4,
                                         dns_nameservers=[],
                                         cidr="192.168.0.0/24")
        flip_model = None
        if flip:
            flip_model = models.IPAddress()
            flip_model.update(flip)
            flip_model.subnet = subnet_model
            if curr_port_model:
                flip_model.ports = [curr_port_model]
            fixed_ip = flip.get("fixed_ip_address")
            if fixed_ip:
                addr = netaddr.IPAddress(fixed_ip)
                fixed_ip_model = models.IPAddress()
                fixed_ip_model.update(dict(address_readable=fixed_ip,
                                           address=int(addr), version=4,
                                           address_type="fixed"))
                flip_model.fixed_ip = fixed_ip_model

        net_model = models.Network()
        if network:
            net_model.update(network)
        net_model.ports = [new_port_model]

        def _find_port(context, id, **kwargs):
            return (curr_port_model if (curr_port_model and
                                        id == curr_port_model.id)
                    else new_port_model)

        def _flip_assoc(context, addr, fixed_ip):
            addr.fixed_ips.append(fixed_ip)
            return addr

        def _flip_disassoc(context, addr):
            addr.fixed_ip = None
            return addr

        def _port_assoc(context, ports, addr, enable_ports=None):
            addr.ports = ports
            return addr

        def _port_dessoc(context, ports, addr):
            addr.associations = []
            addr.ports = []
            return addr

        def mock_notify(context, notif_type, flip):
            """We don't want to notify from tests"""
            pass

        with contextlib.nested(
            mock.patch("quark.db.api.floating_ip_find"),
            mock.patch("quark.db.api.network_find"),
            mock.patch("quark.db.api.subnet_find"),
            mock.patch("quark.db.api.port_find"),
            mock.patch("quark.drivers.unicorn_driver.UnicornDriver"
                       ".register_floating_ip"),
            mock.patch("quark.drivers.unicorn_driver.UnicornDriver"
                       ".update_floating_ip"),
            mock.patch("quark.drivers.unicorn_driver.UnicornDriver"
                       ".remove_floating_ip"),
            mock.patch("quark.db.api.port_associate_ip"),
            mock.patch("quark.db.api.port_disassociate_ip"),
            mock.patch("quark.db.api.floating_ip_associate_fixed_ip"),
            mock.patch("quark.db.api.floating_ip_disassociate_fixed_ip"),
            mock.patch("quark.billing.notify")
        ) as (flip_find, net_find, subnet_find, port_find, reg_flip,
                update_flip, rem_flip, port_assoc, port_dessoc, flip_assoc,
                flip_dessoc, notify):
            flip_find.return_value = flip_model
            net_find.return_value = net_model
            subnet_find.return_value = subnet_model
            port_find.side_effect = _find_port
            port_assoc.side_effect = _port_assoc
            port_dessoc.side_effect = _port_dessoc
            flip_assoc.side_effect = _flip_assoc
            flip_dessoc.side_effect = _flip_disassoc
            notify.side_effect = mock_notify
            # We'll yield a notify to check how many times and with which
            # arguments it was called.
            yield notify

    def test_update_with_conflicted_fixed_ip_and_gateway(self):
        floating_ip_addr = netaddr.IPAddress("10.0.0.1")
        flip = dict(id=1, address=int(floating_ip_addr), version=4,
                    address_readable=str(floating_ip_addr), subnet_id=1,
                    network_id=2, used_by_tenant_id=1)
        network = dict(id="00000000-0000-0000-0000-000000000000",
                       ipam_strategy="ANY")
        subnet = dict(id=1, network_id=1, name=1,
                      tenant_id=1, ip_version=4,
                      routes=[models.Route(cidr='0.0.0.0',
                                           gateway='192.168.0.1')],
                      cidr="192.168.0.0/24",
                      dns_nameservers=[],
                      enable_dhcp=None)

        fixed_ip_addr = netaddr.IPAddress("192.168.0.1")
        fixed_ips = [dict(address_type="fixed", address=int(fixed_ip_addr),
                          version=4, address_readable=str(fixed_ip_addr),
                          allocated_at=datetime.datetime.now())]
        new_port = dict(id="abcdefgh-1111-2222-3333-1234567890ab")

        with self._stubs(flip=flip, new_port=new_port,
                         ips=fixed_ips, subnet=subnet, network=network):
            with self.assertRaises(
                    q_ex.FixedIpAllocatedToGatewayIp):
                content = dict(port_id=new_port["id"])
                self.plugin.update_floatingip(self.context, flip["id"],
                                              dict(floatingip=content))

    def test_update_with_new_port_and_no_previous_port(self):
        new_port = dict(id="2")

        fixed_ip_addr = netaddr.IPAddress("192.168.0.1")
        fixed_ip = dict(address_type="fixed", version=4,
                        address=int(fixed_ip_addr),
                        address_readable=str(fixed_ip_addr),
                        allocated_at=datetime.datetime.now())

        ips = [fixed_ip]

        addr = netaddr.IPAddress("10.0.0.1")
        flip = dict(id="3", fixed_ip_address="172.16.1.1", address=int(addr),
                    address_readable=str(addr))

        with self._stubs(flip=flip, new_port=new_port, ips=ips) as notify:
            content = dict(port_id=new_port["id"])
            ret = self.plugin.update_floatingip(self.context, flip["id"],
                                                dict(floatingip=content))
            self.assertEqual(ret["fixed_ip_address"], "192.168.0.1")
            self.assertEqual(ret["port_id"], new_port["id"])
            notify.assert_called_once_with(self.context, IP_ASSOC,
                                           mock.ANY)

    def test_update_with_new_port(self):
        curr_port = dict(id="1")
        new_port = dict(id="2")
        fixed_ip_addr = netaddr.IPAddress("192.168.0.1")
        fixed_ip = dict(address_type="fixed", version=4,
                        address=int(fixed_ip_addr),
                        address_readable=str(fixed_ip_addr),
                        allocated_at=datetime.datetime.now())

        ips = [fixed_ip]

        addr = netaddr.IPAddress("10.0.0.1")
        flip = dict(id="3", fixed_ip_address="172.16.1.1", address=int(addr),
                    address_readable=str(addr))

        with self._stubs(flip=flip, curr_port=curr_port,
                         new_port=new_port, ips=ips) as notify:
            content = dict(port_id=new_port["id"])
            ret = self.plugin.update_floatingip(self.context, flip["id"],
                                                dict(floatingip=content))
            self.assertEqual(ret["fixed_ip_address"], "192.168.0.1")
            self.assertEqual(ret["port_id"], new_port["id"])
            self.assertEqual(notify.call_count, 2, 'Should notify twice here')
            call_list = [mock.call(self.context, IP_DISASSOC, mock.ANY),
                         mock.call(self.context, IP_ASSOC, mock.ANY)]
            notify.assert_has_calls(call_list, any_order=True)

    def test_update_with_no_port(self):
        curr_port = dict(id="1")
        addr = netaddr.IPAddress("10.0.0.1")
        flip = dict(id="3", fixed_ip_address="172.16.1.1", address=int(addr),
                    address_readable=str(addr))

        with self._stubs(flip=flip, curr_port=curr_port) as notify:
            content = dict(port_id=None)
            ret = self.plugin.update_floatingip(self.context, flip["id"],
                                                dict(floatingip=content))
            self.assertEqual(ret.get("fixed_ip_address"), None)
            self.assertEqual(ret.get("port_id"), None)
            notify.assert_called_once_with(self.context, IP_DISASSOC,
                                           mock.ANY)

    def test_update_with_non_existent_port_should_fail(self):
        addr = netaddr.IPAddress("10.0.0.1")
        flip = dict(id="3", fixed_ip_address="172.16.1.1", address=int(addr),
                    address_readable=str(addr))

        with self._stubs(flip=flip):
            with self.assertRaises(ex.PortNotFound):
                content = dict(port_id="123")
                self.plugin.update_floatingip(self.context, flip["id"],
                                              dict(floatingip=content))

    def test_update_with_port_with_no_fixed_ip_avail_should_fail(self):
        new_port = dict(id="123")
        addr = netaddr.IPAddress("10.0.0.1")
        flip = dict(id="3", fixed_ip_address="172.16.1.1", address=int(addr),
                    address_readable=str(addr))

        with self._stubs(flip=flip, new_port=new_port):
            with self.assertRaises(q_ex.NoAvailableFixedIpsForPort):
                content = dict(port_id="123")
                self.plugin.update_floatingip(self.context, flip["id"],
                                              dict(floatingip=content))

    def test_update_with_same_port_should_fail(self):
        new_port = dict(id="123")
        curr_port = dict(id="123")
        addr = netaddr.IPAddress("10.0.0.1")
        flip = dict(id="3", fixed_ip_address="172.16.1.1", address=int(addr),
                    address_readable=str(addr))

        with self._stubs(flip=flip, new_port=new_port, curr_port=curr_port):
            with self.assertRaises(q_ex.PortAlreadyAssociatedToFloatingIp):
                content = dict(port_id="123")
                self.plugin.update_floatingip(self.context, flip["id"],
                                              dict(floatingip=content))

    def test_update_when_port_has_a_different_flip_should_fail(self):
        new_port = dict(id="123")
        floating_ip_addr = netaddr.IPAddress("192.168.0.1")
        floating_ip = dict(address_type="floating", version=4,
                           address=int(floating_ip_addr),
                           address_readable=str(floating_ip_addr),
                           allocated_at=datetime.datetime.now())

        ips = [floating_ip]

        curr_port = dict(id="456")
        addr = netaddr.IPAddress("10.0.0.1")
        flip = dict(id="3", fixed_ip_address="172.16.1.1", address=int(addr),
                    address_readable=str(addr))

        with self._stubs(flip=flip, new_port=new_port,
                         curr_port=curr_port, ips=ips):
            with self.assertRaises(q_ex.PortAlreadyContainsFloatingIp):
                content = dict(port_id="123")
                self.plugin.update_floatingip(self.context, flip["id"],
                                              dict(floatingip=content))

    def test_update_with_no_port_and_no_previous_port_should_fail(self):
        addr = netaddr.IPAddress("10.0.0.1")
        flip = dict(id="3", fixed_ip_address="172.16.1.1", address=int(addr),
                    address_readable=str(addr))

        with self._stubs(flip=flip):
            with self.assertRaises(q_ex.FloatingIpUpdateNoPortIdSupplied):
                content = dict(port_id=None)
                self.plugin.update_floatingip(self.context, flip["id"],
                                              dict(floatingip=content))

    def test_update_with_missing_port_id_param_should_fail(self):
        with self._stubs():
            with self.assertRaises(ex.BadRequest):
                content = {}
                self.plugin.update_floatingip(self.context, "123",
                                              dict(floatingip=content))



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from quark.drivers import registry
from quark.drivers import registry_base
from quark.tests import test_base


class FakeBaseRegistry(registry_base.DriverRegistryBase):

    def __init__(self):
        self.drivers = {"test_driver_1": 1,
                        "test_driver_2": 2}


class FakeNetDriverRegistry(registry.DriverRegistry):

    def __init__(self):
        self.drivers = {"test_driver_1": 1,
                        "test_driver_2": 2,
                        "test_driver_3": 3}

        self.port_driver_compat_map = {
            "test_driver_2": ["test_driver_1"],
            "test_driver_3": ["test_driver_1",
                              "test_driver_2"]
        }


class TestRegistryBase(test_base.TestBase):

    def setUp(self):
        self.registry = FakeBaseRegistry()

    def test_get_valid(self):
        driver = self.registry.get_driver("test_driver_1")
        self.assertEqual(driver, 1)

        driver = self.registry.get_driver("test_driver_2")
        self.assertEqual(driver, 2)

    def test_get_invalid(self):
        exc = "Driver does_not_exist is not registered."
        with self.assertRaisesRegexp(Exception, exc):
            self.registry.get_driver("does_not_exist")


class TestDriverRegistry(TestRegistryBase):

    def setUp(self):
        self.registry = FakeNetDriverRegistry()

    def test_get_port_driver(self):
        driver = self.registry.get_driver(
            "test_driver_1", port_driver="test_driver_1")
        self.assertEqual(driver, 1)

        driver = self.registry.get_driver(
            "test_driver_2", port_driver="test_driver_2")
        self.assertEqual(driver, 2)

    def test_get_invalid_port_driver(self):
        exc = "Driver does_not_exist is not registered."
        with self.assertRaisesRegexp(Exception, exc):
            self.registry.get_driver(
                "test_driver_1", port_driver="does_not_exist")

    def test_get_compatable_port_driver(self):
        driver = self.registry.get_driver(
            "test_driver_1", port_driver="test_driver_2")
        self.assertEqual(driver, 2)

        driver = self.registry.get_driver(
            "test_driver_1", port_driver="test_driver_3")
        self.assertEqual(driver, 3)

        driver = self.registry.get_driver(
            "test_driver_2", port_driver="test_driver_3")
        self.assertEqual(driver, 3)

    def test_get_incompatable_port_driver(self):
        exc = ("Port driver test_driver_2 not allowed for "
               "underlying network driver test_driver_3.")
        with self.assertRaisesRegexp(Exception, exc):
            self.registry.get_driver(
                "test_driver_3", port_driver="test_driver_2")



"""remove new subnet index make new mac

Revision ID: a0798b3b7418
Revises: 80419263930a
Create Date: 2016-02-09 19:26:27.899610

"""

# revision identifiers, used by Alembic.
revision = 'a0798b3b7418'
down_revision = '80419263930a'

from alembic import op


def upgrade():
    op.drop_index(op.f('ix_quark_subnets_segement_network_version'),
                  table_name='quark_subnets')
    op.drop_index(op.f('ix_quark_mac_addresses_reallocation'),
                  table_name='quark_mac_addresses')
    op.create_index(op.f('ix_quark_mac_addresses_reallocation'),
                    'quark_mac_addresses', ['deallocated', 'deallocated_at'],
                    unique=False)


def downgrade():
    op.drop_index(op.f('ix_quark_mac_addresses_reallocation'),
                  table_name='quark_mac_addresses')
    op.create_index(op.f('ix_quark_mac_addresses_reallocation'),
                    'quark_mac_addresses', ['deallocated_at', 'deallocated'],
                    unique=False)
    op.create_index(op.f('ix_quark_subnets_segement_network_version'),
                    'quark_subnets', ['segment_id', 'network_id',
                                      'ip_version', 'do_not_use'],
                    unique=False)



"""Add transaction_id to quark_mac_addresses

Revision ID: 356d6c0623c8
Revises: 5632aa202d89
Create Date: 2015-04-14 01:53:48.233241

"""

# revision identifiers, used by Alembic.
revision = '356d6c0623c8'
down_revision = '5632aa202d89'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('quark_mac_addresses',
                  sa.Column('transaction_id', sa.Integer(), nullable=True))
    op.create_foreign_key(
        'fk_quark_macs_transaction_id',
        'quark_mac_addresses', 'quark_transactions',
        ['transaction_id'], ['id'])


def downgrade():
    op.drop_constraint(
        'fk_quark_macs_transaction_id', 'quark_mac_addresses',
        type_='foreignkey')
    op.drop_column('quark_mac_addresses', 'transaction_id')



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
#  under the License.

import contextlib

import mock
import netaddr
from neutron_lib import exceptions as n_exc

from quark.db import api as db_api
from quark import exceptions as q_exc
from quark.plugin_modules import mac_address_ranges
from quark.tests import test_quark_plugin


class TestQuarkGetMacAddressRanges(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, mac_range):
        db_mod = "quark.db.api"
        old_context = self.context
        self.context = self.context.elevated()
        with mock.patch("%s.mac_address_range_find" % db_mod) as mar_find:
            mar_find.return_value = mac_range
            yield
        self.context = old_context

    def test_find_mac_ranges(self):
        mar = dict(id=1, cidr="AA:BB:CC/24")
        with self._stubs([mar]):
            res = self.plugin.get_mac_address_ranges(self.context)
            self.assertEqual(res[0]["id"], mar["id"])
            self.assertEqual(res[0]["cidr"], mar["cidr"])

    def test_find_mac_range(self):
        mar = dict(id=1, cidr="AA:BB:CC/24")
        with self._stubs(mar):
            res = self.plugin.get_mac_address_range(self.context, 1)
            self.assertEqual(res["id"], mar["id"])
            self.assertEqual(res["cidr"], mar["cidr"])

    def test_find_mac_range_fail(self):
        with self._stubs(None):
            with self.assertRaises(q_exc.MacAddressRangeNotFound):
                self.plugin.get_mac_address_range(self.context, 1)


class TestQuarkCreateMacAddressRanges(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, mac_range):
        db_mod = "quark.db.api"
        old_context = self.context
        self.context = self.context.elevated()
        with mock.patch("%s.mac_address_range_create" % db_mod) as mar_create:
            mar_create.return_value = mac_range
            yield
        self.context = old_context

    def test_create_range(self):
        mar = dict(mac_address_range=dict(id=1, cidr="AA:BB:CC/24"))
        with self._stubs(mar["mac_address_range"]):
            res = self.plugin.create_mac_address_range(self.context, mar)
            self.assertEqual(res["id"], mar["mac_address_range"]["id"])
            self.assertEqual(res["cidr"], mar["mac_address_range"]["cidr"])

    def test_to_mac_range_cidr_format(self):
        cidr, first, last = mac_address_ranges._to_mac_range("AA:BB:CC/24")
        first_mac = str(netaddr.EUI(first, dialect=netaddr.mac_unix))
        last_mac = str(netaddr.EUI(last, dialect=netaddr.mac_unix))
        self.assertEqual(cidr, "AA:BB:CC:00:00:00/24")
        self.assertEqual(first_mac, "aa:bb:cc:0:0:0")
        self.assertEqual(last_mac, "aa:bb:cd:0:0:0")

    def test_to_mac_range_just_prefix(self):
        cidr, first, last = mac_address_ranges._to_mac_range("AA:BB:CC")
        first_mac = str(netaddr.EUI(first, dialect=netaddr.mac_unix))
        last_mac = str(netaddr.EUI(last, dialect=netaddr.mac_unix))
        self.assertEqual(cidr, "AA:BB:CC:00:00:00/24")
        self.assertEqual(first_mac, "aa:bb:cc:0:0:0")
        self.assertEqual(last_mac, "aa:bb:cd:0:0:0")

    def test_to_mac_range_unix_format(self):
        cidr, first, last = mac_address_ranges._to_mac_range("AA-BB-CC")
        first_mac = str(netaddr.EUI(first, dialect=netaddr.mac_unix))
        last_mac = str(netaddr.EUI(last, dialect=netaddr.mac_unix))
        self.assertEqual(cidr, "AA:BB:CC:00:00:00/24")
        self.assertEqual(first_mac, "aa:bb:cc:0:0:0")
        self.assertEqual(last_mac, "aa:bb:cd:0:0:0")

    def test_to_mac_range_unix_cidr_format(self):
        cidr, first, last = mac_address_ranges._to_mac_range("AA-BB-CC/24")
        first_mac = str(netaddr.EUI(first, dialect=netaddr.mac_unix))
        last_mac = str(netaddr.EUI(last, dialect=netaddr.mac_unix))
        self.assertEqual(cidr, "AA:BB:CC:00:00:00/24")
        self.assertEqual(first_mac, "aa:bb:cc:0:0:0")
        self.assertEqual(last_mac, "aa:bb:cd:0:0:0")

    def test_to_mac_range_unix_cidr_format_normal_length(self):
        cidr, first, last = mac_address_ranges._to_mac_range("aabbcc000000/29")
        first_mac = str(netaddr.EUI(first, dialect=netaddr.mac_unix))
        last_mac = str(netaddr.EUI(last, dialect=netaddr.mac_unix))
        self.assertEqual(cidr, "AA:BB:CC:00:00:00/29")
        self.assertEqual(first_mac, "aa:bb:cc:0:0:0")
        self.assertEqual(last_mac, "aa:bb:cc:8:0:0")

    def test_to_mac_prefix_too_short_fails(self):
        with self.assertRaises(q_exc.InvalidMacAddressRange):
            cidr, first, last = mac_address_ranges._to_mac_range("AA-BB")

    def test_to_mac_prefix_too_long_fails(self):
        with self.assertRaises(q_exc.InvalidMacAddressRange):
            cidr, first, last = mac_address_ranges._to_mac_range(
                "AA-BB-CC-DD-EE-F0-00")

    def test_to_mac_prefix_is_garbage_fails(self):
        with self.assertRaises(q_exc.InvalidMacAddressRange):
            cidr, first, last = mac_address_ranges._to_mac_range("F0-0-BAR")

    def test_create_range_with_do_not_use(self):
        mar = dict(mac_address_range=dict(id=1, cidr="AA:BB:CC/24",
                                          do_not_use=True))
        admin_ctxt = self.context.elevated()
        res = self.plugin.create_mac_address_range(admin_ctxt, mar)
        self.assertEqual(res["cidr"], "AA:BB:CC:00:00:00/24")
        mac_range = db_api.mac_address_range_find(admin_ctxt,
                                                  id=res["id"]).first()
        self.assertTrue(mac_range["do_not_use"])


class TestQuarkDeleteMacAddressRanges(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, mac_range):
        db_mod = "quark.db.api"
        old_context = self.context
        self.context = self.context.elevated()
        with contextlib.nested(
            mock.patch("%s.mac_address_range_find" % db_mod),
            mock.patch("%s.mac_address_range_delete" % db_mod),
        ) as (mar_find, mar_delete):
            mar_find.return_value = mac_range
            yield mar_delete
        self.context = old_context

    def test_mac_address_range_delete_not_found(self):
        with self._stubs(None):
            with self.assertRaises(q_exc.MacAddressRangeNotFound):
                self.plugin.delete_mac_address_range(self.context, 1)

    def test_mac_address_range_delete_in_use(self):
        mar = mock.MagicMock()
        mar.id = 1
        mar.allocated_macs = 1
        with self._stubs(mar):
            with self.assertRaises(q_exc.MacAddressRangeInUse):
                self.plugin.delete_mac_address_range(self.context, 1)

    def test_mac_address_range_delete_success(self):
        mar = mock.MagicMock()
        mar.id = 1
        mar.allocated_macs = 0
        with self._stubs(mar) as mar_delete:
            resp = self.plugin.delete_mac_address_range(self.context, 1)
            self.assertIsNone(resp)
            mar_delete.assert_called_once_with(self.context, mar)


class TestQuarkMacAddressCRUDNotAdminRaises(test_quark_plugin.TestQuarkPlugin):
    def test_mac_ranges_index_fails(self):
        with self.assertRaises(n_exc.NotAuthorized):
            self.plugin.get_mac_address_ranges(self.context)

    def test_show_mac_range_fails(self):
        with self.assertRaises(n_exc.NotAuthorized):
            self.plugin.get_mac_address_range(self.context, 1)

    def test_create_mac_range_fails(self):
        with self.assertRaises(n_exc.NotAuthorized):
            self.plugin.create_mac_address_range(
                self.context, {"mac_address_range": 1})

    def test_delete_mac_range_fails(self):
        with self.assertRaises(n_exc.NotAuthorized):
            self.plugin.delete_mac_address_range(self.context, 1)



"""Added service column to port table

Revision ID: 4dbf83f37bc0
Revises: 33e9e23ba761
Create Date: 2015-05-26 13:27:38.995202

"""

# revision identifiers, used by Alembic.
revision = '4dbf83f37bc0'
down_revision = '1bdc1b574beb'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('quark_ports', sa.Column('service', sa.String(length=255),
                                           nullable=False,
                                           server_default="none"))


def downgrade():
    op.drop_column('quark_ports', 'service')



# Copyright 2016 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Calculations for different cases for additional IP billing
A - allocated
D - deallocated
| - start or end of the billing period, typically 24 hrs

A |     | D
  | A   | D
A | D   |
  | A D |

---IP CURRENTLY ALLOCATED=has tenant_id (send usage at end of period)---
case 1 (allocated before period began, deallocated after period ended)
  send usage as 24 hours
case 2 (allocated during period, deallocated after period ended)
  send usage as timedelta(period_end - allocated_at)

---IP DEALLOCATED DURING PERIOD (send usage at deallocation time)---
case 3 (allocated before period began, deallocated during period)
  send usage as timedelta(deallocated_at - period_start)
case 4 (allocated during period, deallocated during period)
  send usage as timedelta(deallocated_at - allocated_at)

NOTE: notifications are issued at these times:
    case1 and case2 - this script runs nightly and processes the entries
    case3 and case4 - the notifications are sent immediately when the ip
                      is deallocated
NOTE: assumes that the beginning of a billing cycle is midnight.
"""

import datetime
from neutron.common import rpc as n_rpc
from oslo_config import cfg
from oslo_log import log as logging
from sqlalchemy import and_, or_, null
from quark.db import models
from quark import environment as env

from quark import network_strategy

LOG = logging.getLogger(__name__)
CONF = cfg.CONF

quark_opts = [
    cfg.StrOpt('notify_ip_add', default=False,
               help=_("Sends notifications when IP is added")),
    cfg.StrOpt('notify_ip_delete', default=False,
               help=_("Sends notifications when IP is deleted")),
    cfg.StrOpt('notify_flip_associate', default=False,
               help=_("Sends notifications when FLIP is associated")),
    cfg.StrOpt('notify_flip_disassociate', default=False,
               help=_("Sends notifications when FLIP is disassociated")),
    cfg.StrOpt('notify_ip_exists', default=False,
               help=_("Sends 'ip.exists' notifications"))
]
CONF.register_opts(quark_opts, 'QUARK')

PUBLIC_NETWORK_ID = network_strategy.STRATEGY.get_public_net_id()

IP_ADD = 'ip.add'
IP_DEL = 'ip.delete'
IP_ASSOC = 'ip.associate'
IP_DISASSOC = 'ip.disassociate'
IP_EXISTS = 'ip.exists'


def do_notify(context, event_type, payload):
    """Generic Notifier.

    Parameters:
        - `context`: session context
        - `event_type`: the event type to report, i.e. ip.usage
        - `payload`: dict containing the payload to send
    """
    LOG.debug('IP_BILL: notifying {}'.format(payload))

    notifier = n_rpc.get_notifier('network')
    notifier.info(context, event_type, payload)


@env.has_capability(env.Capabilities.IP_BILLING)
def notify(context, event_type, ipaddress, send_usage=False, *args, **kwargs):
    """Method to send notifications.

    We must send USAGE when a public IPv4 address is deallocated or a FLIP is
    associated.
    Parameters:
        - `context`: the context for notifier
        - `event_type`: the event type for IP allocate, deallocate, associate,
        disassociate
        - `ipaddress`: the ipaddress object to notify about
    Returns:
        nothing
    Notes: this may live in the billing module
    """
    if (event_type == IP_ADD and not CONF.QUARK.notify_ip_add) or \
       (event_type == IP_DEL and not CONF.QUARK.notify_ip_delete) or \
       (event_type == IP_ASSOC and not CONF.QUARK.notify_flip_associate) or \
       (event_type == IP_DISASSOC and not CONF.QUARK.notify_flip_disassociate)\
       or (event_type == IP_EXISTS and not CONF.QUARK.notify_ip_exists):
        LOG.debug('IP_BILL: notification {} is disabled by config'.
                  format(event_type))
        return

    # Do not send notifications when we are undoing due to an error
    if 'rollback' in kwargs and kwargs['rollback']:
        LOG.debug('IP_BILL: not sending notification because we are in undo')
        return

    # ip.add needs the allocated_at time.
    # All other events need the current time.
    ts = ipaddress.allocated_at if event_type == IP_ADD else _now()
    payload = build_payload(ipaddress, event_type, event_time=ts)

    # Send the notification with the payload
    do_notify(context, event_type, payload)

    # When we deallocate an IP or associate a FLIP we must send
    # a usage message to billing.
    # In other words when we supply end_time we must send USAGE to billing
    # immediately.
    # Our billing period is 24 hrs. If the address was allocated after midnight
    # send the start_time as as. If the address was allocated yesterday, then
    # send midnight as the start_time.
    # Note: if allocated_at is empty we assume today's midnight.
    if send_usage:
        if ipaddress.allocated_at is not None and \
           ipaddress.allocated_at >= _midnight_today():
            start_time = ipaddress.allocated_at
        else:
            start_time = _midnight_today()
        payload = build_payload(ipaddress,
                                IP_EXISTS,
                                start_time=start_time,
                                end_time=ts)
        do_notify(context, IP_EXISTS, payload)


def build_payload(ipaddress,
                  event_type,
                  event_time=None,
                  start_time=None,
                  end_time=None):
    """Method builds a payload out of the passed arguments.

    Parameters:
        `ipaddress`: the models.IPAddress object
        `event_type`: USAGE,CREATE,DELETE,SUSPEND,or UNSUSPEND
        `start_time`: startTime for cloudfeeds
        `end_time`: endTime for cloudfeeds
    Returns a dictionary suitable to notify billing.
    Message types mapping to cloud feeds for references:
        ip.exists       - USAGE
        ip.add          - CREATE
        ip.delete       - DELETE
        ip.associate    - UP
        ip.disassociate  - DOWN
    Refer to: http://rax.io/cf-api for more details.
    """
    # This is the common part of all message types
    payload = {
        'event_type': unicode(event_type),
        'tenant_id': unicode(ipaddress.used_by_tenant_id),
        'ip_address': unicode(ipaddress.address_readable),
        'ip_version': int(ipaddress.version),
        'ip_type': unicode(ipaddress.address_type),
        'id': unicode(ipaddress.id)
    }

    # Depending on the message type add the appropriate fields
    if event_type == IP_EXISTS:
        if start_time is None or end_time is None:
            raise ValueError('IP_BILL: {} start_time/end_time cannot be empty'
                             .format(event_type))
        payload.update({
            'startTime': unicode(convert_timestamp(start_time)),
            'endTime': unicode(convert_timestamp(end_time))
        })
    elif event_type in [IP_ADD, IP_DEL, IP_ASSOC, IP_DISASSOC]:
        if event_time is None:
            raise ValueError('IP_BILL: {}: event_time cannot be NULL'
                             .format(event_type))
        payload.update({
            'eventTime': unicode(convert_timestamp(event_time)),
            'subnet_id': unicode(ipaddress.subnet_id),
            'network_id': unicode(ipaddress.network_id),
            'public': True if ipaddress.network_id == PUBLIC_NETWORK_ID
            else False,
        })
    else:
        raise ValueError('IP_BILL: bad event_type: {}'.format(event_type))

    return payload


def build_full_day_ips(query, period_start, period_end):
    """Method to build an IP list for the case 1

    when the IP was allocated before the period start
    and is still allocated after the period end.
    This method only looks at public IPv4 addresses.
    """
    # Filter out only IPv4 that have not been deallocated
    ip_list = query.\
        filter(models.IPAddress.version == 4L).\
        filter(models.IPAddress.network_id == PUBLIC_NETWORK_ID).\
        filter(models.IPAddress.used_by_tenant_id is not None).\
        filter(models.IPAddress.allocated_at != null()).\
        filter(models.IPAddress.allocated_at < period_start).\
        filter(or_(models.IPAddress._deallocated is False,
                   models.IPAddress.deallocated_at == null(),
                   models.IPAddress.deallocated_at >= period_end)).all()

    return ip_list


def build_partial_day_ips(query, period_start, period_end):
    """Method to build an IP list for the case 2

    when the IP was allocated after the period start and
    is still allocated after the period end.
    This method only looks at public IPv4 addresses.
    """
    # Filter out only IPv4 that were allocated after the period start
    # and have not been deallocated before the period end.
    # allocated_at will be set to a date
    ip_list = query.\
        filter(models.IPAddress.version == 4L).\
        filter(models.IPAddress.network_id == PUBLIC_NETWORK_ID).\
        filter(models.IPAddress.used_by_tenant_id is not None).\
        filter(and_(models.IPAddress.allocated_at != null(),
                    models.IPAddress.allocated_at >= period_start,
                    models.IPAddress.allocated_at < period_end)).\
        filter(or_(models.IPAddress._deallocated is False,
                   models.IPAddress.deallocated_at == null(),
                   models.IPAddress.deallocated_at >= period_end)).all()

    return ip_list


def calc_periods(hour=0, minute=0):
    """Returns a tuple of start_period and end_period.

    Assumes that the period is 24-hrs.
    Parameters:
        - `hour`: the hour from 0 to 23 when the period ends
        - `minute`: the minute from 0 to 59 when the period ends
    This method will calculate the end of the period as the closest hour/minute
    going backwards.
    It will also calculate the start of the period as the passed hour/minute
    but 24 hrs ago.
    Example, if we pass 0, 0 - we will get the events from 0:00 midnight of the
    day before yesterday until today's midnight.
    If we pass 2,0 - we will get the start time as 2am of the previous morning
    till 2am of today's morning.
    By default it's midnight.
    """
    # Calculate the time intervals in a usable form
    period_end = datetime.datetime.utcnow().replace(hour=hour,
                                                    minute=minute,
                                                    second=0,
                                                    microsecond=0)
    period_start = period_end - datetime.timedelta(days=1)

    # period end should be slightly before the midnight.
    # hence, we subtract a second
    # this will force period_end to store something like:
    # datetime.datetime(2016, 5, 19, 23, 59, 59, 999999)
    # instead of:
    # datetime.datetime(2016, 5, 20,  0,  0,  0,      0)
    period_end -= datetime.timedelta(seconds=1)

    return (period_start, period_end)


def _midnight_today():
    return datetime.datetime.utcnow().replace(hour=0,
                                              minute=0,
                                              second=0)


def convert_timestamp(ts):
    """Converts the timestamp to a format suitable for Billing.

    Examples of a good timestamp for startTime, endTime, and eventTime:
        '2016-05-20T00:00:00Z'
    We must drop microseconds so that Yagi does not get upset.
    Note the trailing 'Z'. Python does not add the 'Z' so we tack it on
    ourselves.
    """
    return ts.replace(microsecond=0).isoformat() + 'Z'


def _now():
    """Method to get the utcnow without microseconds"""
    return datetime.datetime.utcnow().replace(microsecond=0)



# Copyright 2014 Rackspace Hosting Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for# the specific language governing permissions and limitations
#  under the License.

import mock
import netaddr

import contextlib

from neutron import context
from neutron_lib import exceptions as n_exc
from oslo_config import cfg
from oslo_log import log as logging

from quark.db import api as db_api
from quark.db import ip_types
from quark import exceptions as q_exc
import quark.ipam
from quark import plugin
import quark.plugin_modules.ip_addresses as ip_api
import quark.plugin_modules.mac_address_ranges as macrng_api
import quark.plugin_modules.networks as network_api
import quark.plugin_modules.ports as port_api
import quark.plugin_modules.subnets as subnet_api
from quark.tests.functional.base import BaseFunctionalTest

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class QuarkSharedIPs(BaseFunctionalTest):
    def __init__(self, *args, **kwargs):
        super(QuarkSharedIPs, self).__init__(*args, **kwargs)
        self.disassociate_exception = q_exc.PortRequiresDisassociation
        self.cidr = "192.168.2.0/24"
        self.ip_network = netaddr.IPNetwork(self.cidr)
        network = dict(name="public", tenant_id="fake", network_plugin="BASE")
        self.network = {"network": network}
        subnet = dict(ip_version=4, next_auto_assign_ip=2,
                      cidr=self.cidr, first_ip=self.ip_network.first,
                      last_ip=self.ip_network.last, ip_policy=None,
                      tenant_id="fake")
        self.subnet = {"subnet": subnet}
        port1 = {'port': dict(device_id='a')}
        port2 = {'port': dict(device_id='b')}
        port3 = {'port': dict(device_id='c')}
        port4 = {'port': dict(device_id='d')}
        self.ports_info2 = [port1, port2]
        self.ports_info4 = [port1, port2, port3, port4]

    def setUp(self):
        super(QuarkSharedIPs, self).setUp()

    def tearDown(self):
        super(QuarkSharedIPs, self).tearDown()

    def _make_port_body(self, service):
        body = dict(service=service)
        port_info = {"port": dict(body)}
        return port_info

    @contextlib.contextmanager
    def _stubs(self, network_info, subnet_info, ports_info):
        self.ipam = quark.ipam.QuarkIpamANY()
        with contextlib.nested(
                mock.patch("neutron.common.rpc.get_notifier"),
                mock.patch("neutron.quota.QUOTAS.limit_check")):
            net = network_api.create_network(self.admin_context, network_info)
            mac = {'mac_address_range': dict(cidr="AA:BB:CC")}
            macrng_api.create_mac_address_range(self.admin_context, mac)
            subnet_info['subnet']['network_id'] = net['id']
            sub = subnet_api.create_subnet(self.context, subnet_info)
            ports = []
            for port_info in ports_info:
                port_info['port']['network_id'] = net['id']
                ports.append(port_api.create_port(self.context, port_info))
            yield net, sub, ports

    def test_delete_ip_with_shared_owner_error(self):

        with self._stubs(self.network, self.subnet, self.ports_info2) as (
                net, sub, ports):

            port_ids = [ports[0]['id'], ports[1]['id']]
            p_id = ports[0]['id']
            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4)}
            ip = ip_api.create_ip_address(self.context, shared_ip)
            port_ip_update = ip_api.update_port_for_ip_address
            port_ip_update(self.context, ip['id'], p_id,
                           self._make_port_body('derp'))

            with self.assertRaises(self.disassociate_exception):
                ip_api.delete_ip_address(self.context, ip['id'])

    def test_update_shared_ip_with_plural_will_error(self):

        with self._stubs(self.network, self.subnet, self.ports_info4) as (
                net, sub, ports):

            port_ids = [ports[0]['id'], ports[1]['id']]
            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4)}
            ip = ip_api.create_ip_address(self.context, shared_ip)
            self.assertEqual(ip_types.SHARED, ip['type'])
            port_ids = [ports[0]['id'], ports[3]['id']]
            shared_ip = {'ip_addresses': dict(port_ids=port_ids)}

            with self.assertRaises(n_exc.BadRequest):
                ip_api.update_ip_address(self.context, ip['id'], shared_ip)

    def test_update_shared_ip_with_empty_port_id_list_will_error(self):

        with self._stubs(self.network, self.subnet, self.ports_info4) as (
                net, sub, ports):

            port_ids = [ports[0]['id'], ports[1]['id']]
            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4)}
            ip = ip_api.create_ip_address(self.context, shared_ip)
            self.assertEqual(ip_types.SHARED, ip['type'])
            port_ids = []
            shared_ip = {'ip_addresses': dict(port_ids=port_ids)}

            with self.assertRaises(n_exc.BadRequest):
                ip_api.update_ip_address(self.context, ip['id'], shared_ip)

    def test_update_shared_ip_with_garbage_will_error(self):

        with self._stubs(self.network, self.subnet, self.ports_info4) as (
                net, sub, ports):

            port_ids = [ports[0]['id'], ports[1]['id']]
            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4)}
            ip = ip_api.create_ip_address(self.context, shared_ip)
            self.assertEqual(ip_types.SHARED, ip['type'])
            port_ids = [ports[0]['id'], ports[3]['id']]
            shared_ip = {'delasdfkj': dict(port_ids=port_ids)}

            with self.assertRaises(n_exc.BadRequest):
                ip_api.update_ip_address(self.context, ip['id'], shared_ip)

    def test_update_shared_ip_with_unowned_ports_is_okay(self):

        with self._stubs(self.network, self.subnet, self.ports_info4) as (
                net, sub, ports):

            port_ids = [ports[0]['id'], ports[1]['id']]
            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4)}
            ip = ip_api.create_ip_address(self.context, shared_ip)
            self.assertEqual(ip_types.SHARED, ip['type'])
            port_ids = [ports[0]['id'], ports[3]['id']]
            shared_ip = {'ip_address': dict(port_ids=port_ids)}

            ip = ip_api.update_ip_address(self.context, ip['id'], shared_ip)

            ports_ip = ip_api.get_ports_for_ip_address(self.context, ip['id'])
            self.assertEqual(2, len(ports_ip))
            for port in ports_ip:
                self.assertTrue(port['id'] in port_ids)

    def test_has_shared_owner_detection(self):
        with self._stubs(self.network, self.subnet, self.ports_info4) as (
                net, sub, ports):

            port_ids = [ports[0]['id'], ports[1]['id']]
            p_id = ports[0]['id']

            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4)}
            ip = ip_api.create_ip_address(self.context, shared_ip)
            self.assertEqual(ip_types.SHARED, ip['type'])
            ip_db = db_api.ip_address_find(self.context, id=ip['id'],
                                           scope=db_api.ONE)
            self.assertFalse(ip_db.has_any_shared_owner())

            port_ip_update = ip_api.update_port_for_ip_address
            port_ip_update(self.context, ip['id'], p_id,
                           self._make_port_body('derp'))

            self.assertTrue(ip_db.has_any_shared_owner())

    def test_update_shared_ip_with_owned_port_no_error_if_present(self):

        with self._stubs(self.network, self.subnet, self.ports_info4) as (
                net, sub, ports):

            port_ids = [ports[0]['id'], ports[1]['id']]
            p_id = ports[0]['id']

            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4)}
            ip = ip_api.create_ip_address(self.context, shared_ip)
            self.assertEqual(ip_types.SHARED, ip['type'])

            port_ip_update = ip_api.update_port_for_ip_address
            port_ip_update(self.context, ip['id'], p_id,
                           self._make_port_body('derp'))

            port_ids = [ports[0]['id'], ports[2]['id']]
            shared_ip = {'ip_address': dict(port_ids=port_ids)}

            ip_api.update_ip_address(self.context, ip['id'], shared_ip)

    def test_update_shared_ip_with_owned_port_no_error_if_present_alone(self):

        with self._stubs(self.network, self.subnet, self.ports_info4) as (
                net, sub, ports):

            port_ids = [ports[0]['id'], ports[1]['id']]
            p_id = ports[0]['id']

            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4)}
            ip = ip_api.create_ip_address(self.context, shared_ip)
            self.assertEqual(ip_types.SHARED, ip['type'])

            port_ip_update = ip_api.update_port_for_ip_address
            port_ip_update(self.context, ip['id'], p_id,
                           self._make_port_body('derp'))

            port_ids = [ports[0]['id']]
            shared_ip = {'ip_address': dict(port_ids=port_ids)}

            ip_api.update_ip_address(self.context, ip['id'], shared_ip)

    def test_update_shared_ip_with_owned_port_no_error_if_adding(self):

        with self._stubs(self.network, self.subnet, self.ports_info4) as (
                net, sub, ports):

            port_ids = [ports[0]['id'], ports[1]['id']]
            p_id = ports[0]['id']

            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4)}
            ip = ip_api.create_ip_address(self.context, shared_ip)
            self.assertEqual(ip_types.SHARED, ip['type'])

            port_ip_update = ip_api.update_port_for_ip_address
            port_ip_update(self.context, ip['id'], p_id,
                           self._make_port_body('derp'))
            port_ids = [ports[0]['id'], ports[1]['id'], ports[2]['id']]

            shared_ip = {'ip_address': dict(port_ids=port_ids)}

            ip_api.update_ip_address(self.context, ip['id'], shared_ip)

    def test_update_shared_ip_with_owned_port_no_error_if_same_list(self):

        with self._stubs(self.network, self.subnet, self.ports_info4) as (
                net, sub, ports):

            port_ids = [ports[0]['id'], ports[1]['id']]
            p_id = ports[0]['id']

            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4)}
            ip = ip_api.create_ip_address(self.context, shared_ip)
            self.assertEqual(ip_types.SHARED, ip['type'])

            port_ip_update = ip_api.update_port_for_ip_address
            port_ip_update(self.context, ip['id'], p_id,
                           self._make_port_body('derp'))

            shared_ip = {'ip_address': dict(port_ids=port_ids)}

            ip_api.update_ip_address(self.context, ip['id'], shared_ip)

    def test_update_shared_ip_with_owned_port_error(self):

        with self._stubs(self.network, self.subnet, self.ports_info4) as (
                net, sub, ports):

            port_ids = [ports[0]['id'], ports[1]['id']]
            p_id = ports[0]['id']

            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4)}
            ip = ip_api.create_ip_address(self.context, shared_ip)
            self.assertEqual(ip_types.SHARED, ip['type'])

            port_ip_update = ip_api.update_port_for_ip_address
            port_ip_update(self.context, ip['id'], p_id,
                           self._make_port_body('derp'))

            port_ids = [ports[2]['id'], ports[1]['id']]
            shared_ip = {'ip_address': dict(port_ids=port_ids)}

            with self.assertRaises(self.disassociate_exception):
                ip_api.update_ip_address(self.context, ip['id'], shared_ip)

    def test_create_shared_ips_with_port_ids(self):

        with self._stubs(self.network, self.subnet, self.ports_info2) as (
                net, sub, ports):

            port_ids = [ports[0]['id'], ports[1]['id']]
            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4)}
            ip = ip_api.create_ip_address(self.context, shared_ip)
            self.assertEqual(ip_types.SHARED, ip['type'])

            ports_ip = ip_api.get_ports_for_ip_address(self.context, ip['id'])
            self.assertEqual(2, len(ports_ip))

    def test_create_shared_ips_without_tenant_id_raises(self):
        no_tenant_context = context.Context(None, None, is_admin=True)

        with self._stubs(self.network, self.subnet, self.ports_info2) as (
                net, sub, ports):

            port_ids = [ports[0]['id'], ports[1]['id']]
            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4)}
            api = plugin.Plugin()
            with self.assertRaises(n_exc.BadRequest):
                api.create_ip_address(no_tenant_context, shared_ip)

    def test_create_shared_ips_fails_with_plural_body(self):

        with self._stubs(self.network, self.subnet, self.ports_info2) as (
                net, sub, ports):

            port_ids = [ports[0]['id'], ports[1]['id']]
            shared_ip = {'ip_addresses': dict(port_ids=port_ids,
                                              network_id=net['id'],
                                              version=4)}
            with self.assertRaises(n_exc.BadRequest):
                ip_api.create_ip_address(self.context, shared_ip)

    def test_create_shared_ips_fails_with_garbage_body(self):

        with self._stubs(self.network, self.subnet, self.ports_info2) as (
                net, sub, ports):

            port_ids = [ports[0]['id'], ports[1]['id']]
            shared_ip = {'derpie_derp': dict(port_ids=port_ids,
                                             network_id=net['id'],
                                             version=4)}
            with self.assertRaises(n_exc.BadRequest):
                ip_api.create_ip_address(self.context, shared_ip)

    def test_shared_ip_in_fixed_ip_list(self):

        with self._stubs(self.network, self.subnet, self.ports_info2) as (
                net, sub, ports):

            for port in ports:
                self.assertEqual(1, len(port['fixed_ips']))

            port_ids = [ports[0]['id'], ports[1]['id']]
            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4)}
            p_id = ports[0]['id']

            ip = ip_api.create_ip_address(self.context, shared_ip)
            self.assertEqual(ip_types.SHARED, ip['type'])

            ports_ip = ip_api.get_ports_for_ip_address(self.context, ip['id'])
            self.assertEqual(2, len(ports_ip))

            port_ip_update = ip_api.update_port_for_ip_address
            updated_port = port_ip_update(self.context, ip['id'],
                                          p_id, self._make_port_body('derp'))
            self.assertEqual('derp', updated_port.get('service'))

            port = port_api.get_port(self.context, p_id)
            self.assertEqual(2, len(port['fixed_ips']))

    def test_ip_port_list_has_services(self):

        with self._stubs(self.network, self.subnet, self.ports_info2) as (
                net, sub, ports):

            for port in ports:
                self.assertEqual(1, len(port['fixed_ips']))

            device_ids = [ports[0]['device_id'], ports[1]['device_id']]
            shared_ip = {'ip_address': dict(device_ids=device_ids,
                                            network_id=net['id'],
                                            version=4)}
            ip = ip_api.create_ip_address(self.context, shared_ip)
            port_ip_update = ip_api.update_port_for_ip_address
            port_ip_update(self.context, ip['id'],
                           ports[0]['id'], self._make_port_body('derp'))

            ports_ip = ip_api.get_ports_for_ip_address(self.context, ip['id'])
            self.assertEqual(2, len(ports_ip))

            for port in ports_ip:
                self.assertTrue('service' in port)
                self.assertTrue('device_id' in port)
                self.assertTrue('id' in port)
                self.assertTrue(port['service'] in ('derp', 'none'),
                                'Service is: %s' % str(port['service']))

    def test_can_delete_ip_without_active_port(self):

        with self._stubs(self.network, self.subnet, self.ports_info2) as (
                net, sub, ports):
            device_ids = [ports[0]['device_id'], ports[1]['device_id']]
            shared_ip = {'ip_address': dict(device_ids=device_ids,
                                            network_id=net['id'],
                                            version=4)}
            ip = ip_api.create_ip_address(self.context, shared_ip)
            ip_api.delete_ip_address(self.context, ip['id'])
            with self.assertRaises(q_exc.IpAddressNotFound):
                ip_api.get_ip_address(self.context, ip['id'])

    def test_cannot_delete_ip_with_active_port(self):

        with self._stubs(self.network, self.subnet, self.ports_info2) as (
                net, sub, ports):
            device_ids = [ports[0]['device_id'], ports[1]['device_id']]
            shared_ip = {'ip_address': dict(device_ids=device_ids,
                                            network_id=net['id'],
                                            version=4)}
            ip = ip_api.create_ip_address(self.context, shared_ip)
            port_ip_update = ip_api.update_port_for_ip_address
            port_ip_update(self.context, ip['id'],
                           ports[0]['id'], self._make_port_body('derp'))

            with self.assertRaises(self.disassociate_exception):
                ip_api.delete_ip_address(self.context, ip['id'])


class QuarkSharedIPsQuotaCheck(BaseFunctionalTest):
    def __init__(self, *args, **kwargs):
        super(QuarkSharedIPsQuotaCheck, self).__init__(*args, **kwargs)
        self.disassociate_exception = q_exc.PortRequiresDisassociation
        self.cidr = "192.168.2.0/24"
        self.ip_network = netaddr.IPNetwork(self.cidr)
        network = dict(name="public", tenant_id="fake", network_plugin="BASE")
        self.network = {"network": network}
        subnet = dict(ip_version=4, next_auto_assign_ip=2,
                      cidr=self.cidr, first_ip=self.ip_network.first,
                      last_ip=self.ip_network.last, ip_policy=None,
                      tenant_id="fake")
        self.subnet = {"subnet": subnet}
        port1 = {'port': dict(device_id='a')}
        port2 = {'port': dict(device_id='b')}
        port3 = {'port': dict(device_id='c')}
        port4 = {'port': dict(device_id='d')}
        self.ports_info2 = [port1, port2]
        self.ports_info4 = [port1, port2, port3, port4]

    def setUp(self):
        super(QuarkSharedIPsQuotaCheck, self).setUp()

    def tearDown(self):
        super(QuarkSharedIPsQuotaCheck, self).tearDown()

    @contextlib.contextmanager
    def _stubs(self, network_info, subnet_info, ports_info):
        self.ipam = quark.ipam.QuarkIpamANY()
        with contextlib.nested(
                mock.patch("neutron.common.rpc.get_notifier")):
            self.context.is_admin = True
            net = network_api.create_network(self.context, network_info)
            mac = {'mac_address_range': dict(cidr="AA:BB:CC")}
            macrng_api.create_mac_address_range(self.context, mac)
            self.context.is_admin = False
            subnet_info['subnet']['network_id'] = net['id']
            sub = subnet_api.create_subnet(self.context, subnet_info)
            ports = []
            for port_info in ports_info:
                port_info['port']['network_id'] = net['id']
                ports.append(port_api.create_port(self.context, port_info))
            yield net, sub, ports

    def test_create_shared_ip_over_public_network_quota(self):
        network = dict(name="public", tenant_id="fake", network_plugin="BASE",
                       id='00000000-0000-0000-0000-000000000000')
        network = {"network": network}

        with self._stubs(network, self.subnet, self.ports_info4) as (
                net, sub, ports):

            port_ids = [ports[0]['id'], ports[1]['id']]
            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4)}
            port_ids = [ports[2]['id'], ports[3]['id']]
            shared_ip2 = {'ip_address': dict(port_ids=port_ids,
                                             network_id=net['id'],
                                             version=4)}

            # NOTE(roaet): this is hardcoded to 5 and will fail after 5
            for i in xrange(5):
                # NOTE(roaet): need to do this modulo stuff to not hit IP quota
                if i % 2 == 0:
                    ip_api.create_ip_address(self.context, shared_ip)
                else:
                    ip_api.create_ip_address(self.context, shared_ip2)

            with self.assertRaises(q_exc.CannotCreateMoreSharedIPs):
                ip_api.create_ip_address(self.context, shared_ip)

    def test_create_shared_ip_over_isolated_network_quota(self):
        with self._stubs(self.network, self.subnet, self.ports_info4) as (
                net, sub, ports):

            port_ids = [ports[0]['id'], ports[1]['id']]
            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4)}
            port_ids = [ports[2]['id'], ports[3]['id']]
            shared_ip2 = {'ip_address': dict(port_ids=port_ids,
                                             network_id=net['id'],
                                             version=4)}

            # NOTE(roaet): this is hardcoded to 5
            for i in xrange(4):
                # NOTE(roaet): need to do this modulo stuff to not hit IP quota
                if i % 2 == 0:
                    ip_api.create_ip_address(self.context, shared_ip)
                else:
                    ip_api.create_ip_address(self.context, shared_ip2)

            # NOTE(roaet): this should not fail
            ip_api.create_ip_address(self.context, shared_ip)

    def test_create_shared_ip_over_service_network_quota(self):
        network = dict(name="service", tenant_id="fake", network_plugin="BASE",
                       id='11111111-1111-1111-1111-111111111111')
        network = {"network": network}

        with self._stubs(network, self.subnet, self.ports_info4) as (
                net, sub, ports):

            port_ids = [ports[0]['id'], ports[1]['id']]
            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4)}

            # NOTE(roaet): this is hardcoded to 0 so should fail instantly
            with self.assertRaises(q_exc.CannotCreateMoreSharedIPs):
                ip_api.create_ip_address(self.context, shared_ip)

    def test_create_shared_ip_over_public_total_ip_on_port_quota(self):
        network = dict(name="public", tenant_id="fake", network_plugin="BASE",
                       id='00000000-0000-0000-0000-000000000000')
        network = {"network": network}

        with self._stubs(network, self.subnet, self.ports_info2) as (
                net, sub, ports):

            port_ids = [ports[0]['id'], ports[1]['id']]
            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4)}
            # NOTE(roaet): this is hardcoded to 6 so should fail after 5
            # since a port comes with 1 IP already
            for i in xrange(5):
                ip_api.create_ip_address(self.context, shared_ip)

            with self.assertRaises(n_exc.OverQuota):
                ip_api.create_ip_address(self.context, shared_ip)

    def test_create_shared_ip_over_isolated_total_ip_on_port_quota(self):
        with self._stubs(self.network, self.subnet, self.ports_info2) as (
                net, sub, ports):

            port_ids = [ports[0]['id'], ports[1]['id']]
            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4)}
            # NOTE(roaet): this is hardcoded to 5 should fail after 4
            # since a port comes with 1 IP already
            for i in xrange(4):
                ip_api.create_ip_address(self.context, shared_ip)

            with self.assertRaises(n_exc.OverQuota):
                ip_api.create_ip_address(self.context, shared_ip)

    def test_create_shared_ip_over_service_total_ip_on_port_quota(self):
        network = dict(name="service", tenant_id="fake", network_plugin="BASE",
                       id='11111111-1111-1111-1111-111111111111')
        network = {"network": network}

        with self._stubs(network, self.subnet, self.ports_info2) as (
                net, sub, ports):

            port_ids = [ports[0]['id'], ports[1]['id']]
            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4)}
            # NOTE(roaet): this is hardcoded to 1 so should fail immediately
            with self.assertRaises(n_exc.OverQuota):
                ip_api.create_ip_address(self.context, shared_ip)


class QuarkTestSharedIpAddressesQuota(BaseFunctionalTest):
    def __init__(self, *args, **kwargs):
        super(QuarkTestSharedIpAddressesQuota, self).__init__(*args, **kwargs)
        self.disassociate_exception = q_exc.PortRequiresDisassociation
        self.cidr = "192.168.2.0/24"
        self.ip_network = netaddr.IPNetwork(self.cidr)
        self.network_id = "55555555-5555-5555-5555-555555555555"
        network = dict(id=self.network_id, name="public", tenant_id="fake",
                       network_plugin="BASE")
        self.network = {"network": network}
        subnet = dict(ip_version=4, next_auto_assign_ip=2,
                      cidr=self.cidr, first_ip=self.ip_network.first,
                      last_ip=self.ip_network.last, ip_policy=None,
                      tenant_id="fake")
        self.subnet = {"subnet": subnet}
        port1 = {'port': dict(device_id='a')}
        port2 = {'port': dict(device_id='b')}
        port3 = {'port': dict(device_id='c')}
        port4 = {'port': dict(device_id='d')}
        port1a = {'port': dict(device_id='x', tenant_id='derp')}
        port2a = {'port': dict(device_id='y', tenant_id='derp')}
        port3a = {'port': dict(device_id='z', tenant_id='derp')}
        self.ports_info2 = [port1, port2]
        self.ports_info2_other = [port1a, port2a, port3a]
        self.ports_info4 = [port1, port2, port3, port4]
        self.shared_ip_on_network = {self.network_id: 3}
        super(QuarkTestSharedIpAddressesQuota, self).setUp()

    def setUp(self):
        super(QuarkTestSharedIpAddressesQuota, self).setUp()
        self.old_cfg = cfg.CONF.QUARK.shared_ips_allowed_on_network
        cfg.CONF.set_override(
            'shared_ips_allowed_on_network', self.shared_ip_on_network,
            "QUARK")

    def tearDown(self):
        super(QuarkTestSharedIpAddressesQuota, self).tearDown()
        cfg.CONF.set_override(
            'shared_ips_allowed_on_network', self.old_cfg,
            "QUARK")

    def _make_port_body(self, service):
        body = dict(service=service)
        port_info = {"port": dict(body)}
        return port_info

    @contextlib.contextmanager
    def _stubs(self, network_info, subnet_info, ports_info):
        self.ipam = quark.ipam.QuarkIpamANY()
        with contextlib.nested(
                mock.patch("neutron.common.rpc.get_notifier"),
                mock.patch("neutron.quota.QUOTAS.limit_check")):
            net = network_api.create_network(self.admin_context, network_info)
            mac = {'mac_address_range': dict(cidr="AA:BB:CC")}
            macrng_api.create_mac_address_range(self.admin_context, mac)
            subnet_info['subnet']['network_id'] = net['id']
            sub = subnet_api.create_subnet(self.context, subnet_info)
            ports = []
            for port_info in ports_info:
                port_info['port']['network_id'] = net['id']
                ports.append(port_api.create_port(self.context, port_info))
            yield net, sub, ports

    def test_create_lots_of_shared_ips_same_tenant_raises(self):
        with self._stubs(self.network, self.subnet, self.ports_info4) as (
                net, sub, ports):

            port_ids = [ports[0]['id'], ports[1]['id']]
            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4)}
            ip = ip_api.create_ip_address(self.context, shared_ip)
            self.assertEqual(ip_types.SHARED, ip['type'])

            port_ids = [ports[1]['id'], ports[2]['id']]
            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4)}
            ip = ip_api.create_ip_address(self.context, shared_ip)
            self.assertEqual(ip_types.SHARED, ip['type'])

            port_ids = [ports[2]['id'], ports[3]['id']]
            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4)}
            ip = ip_api.create_ip_address(self.context, shared_ip)
            self.assertEqual(ip_types.SHARED, ip['type'])

            port_ids = [ports[0]['id'], ports[3]['id']]
            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4)}
            with self.assertRaises(q_exc.CannotCreateMoreSharedIPs):
                ip_api.create_ip_address(self.context, shared_ip)

    def test_create_lots_of_shared_ips_diff_tenant_raises(self):
        with self._stubs(self.network, self.subnet, self.ports_info4) as (
                net, sub, ports):
            derp_ports = []
            derp_context = context.Context("derp", "derp", is_admin=True)
            for port_info in self.ports_info2_other:
                port_info['port']['network_id'] = net['id']
                port = port_api.create_port(derp_context, port_info)
                self.assertEqual("derp", port['tenant_id'])
                derp_ports.append(port)

            port_ids = [derp_ports[0]['id'], derp_ports[1]['id']]
            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4, tenant_id="derp")}
            ip = ip_api.create_ip_address(derp_context, shared_ip)
            self.assertEqual(ip_types.SHARED, ip['type'])
            self.assertEqual("derp", ip['tenant_id'])

            ips = db_api.ip_address_find(self.admin_context,
                                         network_id=net['id'],
                                         address_type=ip_types.SHARED,
                                         scope=db_api.ALL)
            self.assertEqual(1, len(ips))

            port_ids = [ports[1]['id'], ports[2]['id']]
            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4)}
            ip = ip_api.create_ip_address(self.context, shared_ip)
            self.assertEqual(ip_types.SHARED, ip['type'])
            self.assertEqual("fake", ip['tenant_id'])

            ips = db_api.ip_address_find(self.admin_context,
                                         network_id=net['id'],
                                         address_type=ip_types.SHARED,
                                         scope=db_api.ALL)
            self.assertEqual(2, len(ips))

            port_ids = [ports[2]['id'], ports[3]['id']]
            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4)}
            ip = ip_api.create_ip_address(self.context, shared_ip)
            self.assertEqual(ip_types.SHARED, ip['type'])
            self.assertEqual("fake", ip['tenant_id'])

            ips = db_api.ip_address_find(self.admin_context,
                                         network_id=net['id'],
                                         address_type=ip_types.SHARED,
                                         scope=db_api.ALL)
            self.assertEqual(3, len(ips))

            port_ids = [ports[0]['id'], ports[3]['id']]
            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4)}
            ip = ip_api.create_ip_address(self.context, shared_ip)
            self.assertEqual(ip_types.SHARED, ip['type'])
            self.assertEqual("fake", ip['tenant_id'])

            ips = db_api.ip_address_find(self.admin_context,
                                         network_id=net['id'],
                                         address_type=ip_types.SHARED,
                                         scope=db_api.ALL)
            self.assertEqual(4, len(ips))  # total shared IPs on network

            ips = db_api.ip_address_find(self.admin_context,
                                         network_id=net['id'],
                                         tenant_id=self.context.tenant_id,
                                         address_type=ip_types.SHARED,
                                         scope=db_api.ALL)
            self.assertEqual(3, len(ips))

            port_ids = [ports[0]['id'], ports[1]['id']]
            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4)}
            with self.assertRaises(q_exc.CannotCreateMoreSharedIPs):
                ip_api.create_ip_address(self.context, shared_ip)

            port_ids = [derp_ports[1]['id'], derp_ports[2]['id']]
            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4, tenant_id="derp")}
            ip = ip_api.create_ip_address(derp_context, shared_ip)
            self.assertEqual(ip_types.SHARED, ip['type'])
            self.assertEqual("derp", ip['tenant_id'])

            ips = db_api.ip_address_find(self.admin_context,
                                         network_id=net['id'],
                                         address_type=ip_types.SHARED,
                                         scope=db_api.ALL)
            self.assertEqual(5, len(ips))

            ips = db_api.ip_address_find(self.admin_context,
                                         network_id=net['id'],
                                         tenant_id='derp',
                                         address_type=ip_types.SHARED,
                                         scope=db_api.ALL)
            self.assertEqual(2, len(ips))

            port_ids = [derp_ports[0]['id'], derp_ports[2]['id']]
            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4, tenant_id="derp")}
            ip = ip_api.create_ip_address(derp_context, shared_ip)
            self.assertEqual(ip_types.SHARED, ip['type'])
            self.assertEqual("derp", ip['tenant_id'])

            ips = db_api.ip_address_find(self.admin_context,
                                         network_id=net['id'],
                                         address_type=ip_types.SHARED,
                                         scope=db_api.ALL)
            self.assertEqual(6, len(ips))

            ips = db_api.ip_address_find(self.admin_context,
                                         network_id=net['id'],
                                         tenant_id='derp',
                                         address_type=ip_types.SHARED,
                                         scope=db_api.ALL)
            self.assertEqual(3, len(ips))

            port_ids = [derp_ports[0]['id'], derp_ports[2]['id']]
            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4, tenant_id="derp")}
            with self.assertRaises(q_exc.CannotCreateMoreSharedIPs):
                ip_api.create_ip_address(self.admin_context, shared_ip)



"""Truncate IPPolicyCIDR outside of Subnet._cidr

Revision ID: 2748e48cee3a
Revises: 4358d1b8cc75
Create Date: 2014-06-27 05:46:09.430767

"""

# revision identifiers, used by Alembic.
revision = '2748e48cee3a'
down_revision = '1284c81cf727'

import logging

from alembic import op
from sqlalchemy.sql import column, select, table
import netaddr
from oslo_utils import timeutils
from oslo_utils import uuidutils
import sqlalchemy as sa

LOG = logging.getLogger("alembic.migration")


def upgrade():
    ip_policy_cidrs = table('quark_ip_policy_cidrs',
                            column('id', sa.String(length=36)),
                            column('created_at', sa.DateTime()),
                            column('ip_policy_id', sa.String(length=36)),
                            column('cidr', sa.String(length=64)))
    subnets = table('quark_subnets',
                    column('_cidr', sa.String(length=64)),
                    column('ip_policy_id', sa.String(length=36)))

    connection = op.get_bind()

    # 1. Find `quark_ip_policy_cidrs` rows.
    data = connection.execute(select([
        subnets.c.ip_policy_id, subnets.c._cidr,
        ip_policy_cidrs.c.id, ip_policy_cidrs.c.cidr]).where(
            subnets.c.ip_policy_id == ip_policy_cidrs.c.ip_policy_id).order_by(
                subnets.c.ip_policy_id)).fetchall()
    if data is None:
        return

    # 2. Accumulate with `quark_ip_policy_cidrs` rows are outside of the
    #    subnet's cidr.
    ipp_to_update = dict()

    def _test_change_needed(ipp_id, s, ipp):
        if s is None or ipp is None:
            return
        diff = ipp - s
        if diff.size > 0:
            ipp_to_update[ipp_id] = ipp & s

    prev_ip_policy_id = ''
    subnet, ip_policy = None, None
    for ip_policy_id, cidr, ippc_id, ippc_cidr in data:
        if ip_policy_id != prev_ip_policy_id:
            _test_change_needed(prev_ip_policy_id, subnet, ip_policy)
            subnet, ip_policy = netaddr.IPSet([cidr]), netaddr.IPSet()
        ip_policy |= netaddr.IPSet([ippc_cidr])
        prev_ip_policy_id = ip_policy_id
    _test_change_needed(prev_ip_policy_id, subnet, ip_policy)

    if not ipp_to_update.keys():
        return

    LOG.info("IP Policy IDs to update: %s", ipp_to_update.keys())

    # 3. Delete `quark_ip_policy_cidrs` rows that need to be replaced with rows
    #    that are inside of the subnet's cidr.
    connection.execute(ip_policy_cidrs.delete().where(
        ip_policy_cidrs.c.ip_policy_id.in_(ipp_to_update.keys())))

    # 4. Insert `quark_ip_policy_cidrs` rows with cidrs that are inside the
    #    subnet's cidr.
    vals = [dict(id=uuidutils.generate_uuid(),
                 created_at=timeutils.utcnow(),
                 ip_policy_id=key,
                 cidr=str(x.cidr))
            for key in ipp_to_update.keys()
            for x in ipp_to_update[key].iter_cidrs()]
    if not vals:
        return

    LOG.info("IP Policy CIDR IDs to insert: %s", [v["id"] for v in vals])

    connection.execute(ip_policy_cidrs.insert(), *vals)


def downgrade():
    raise NotImplementedError()



# Copyright (c) 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from neutron.common import config
from neutron import context
from oslo_config import cfg
import unittest2


class TestBase(unittest2.TestCase):
    '''Class to decide which unit test class to inherit from uniformly.'''

    def setUp(self):
        super(TestBase, self).setUp()
        tox_path = os.environ.get("VIRTUAL_ENV")
        cfg.CONF.set_override('state_path', tox_path)

        neutron_conf_path = "%s/etc/neutron/neutron.conf" % tox_path
        try:
            open(neutron_conf_path, "r")
        except IOError:
            open(neutron_conf_path, "w")

        args = ['--config-file', neutron_conf_path]
        config.init(args=args)

        self.context = context.Context('fake', 'fake', is_admin=False)
        self.admin_context = context.Context('fake', 'fake', is_admin=True,
                                             load_admin_roles=False)

        class FakeContext(object):
            def __new__(cls, *args, **kwargs):
                return super(FakeContext, cls).__new__(cls)

            def __enter__(*args, **kwargs):
                pass

            def __exit__(*args, **kwargs):
                pass

        self.context.session.begin = FakeContext
